import torch
import torch.nn as nn
import torchvision.transforms as T
from torch.utils.data import DataLoader, Dataset
import numpy as np
import os
from PIL import Image
from datetime import datetime


class VOC_SEG(Dataset):
    def __init__(self, root, width, height, train=True, transforms=None):
        # 图像统一剪切尺寸（width, height）
        self.width = width
        self.height = height
        # VOC数据集中对应的标签
        self.classes = ['background', 'aeroplane', 'bicycle', 'bird', 'boat',
                        'bottle', 'bus', 'car', 'cat', 'chair', 'cow', 'diningtable',
                        'dog', 'horse', 'motorbike', 'person', 'potted plant',
                        'sheep', 'sofa', 'train', 'tv/monitor']
        # 各种标签所对应的颜色
        self.colormap = [[0, 0, 0], [128, 0, 0], [0, 128, 0], [128, 128, 0], [0, 0, 128],
                         [128, 0, 128], [0, 128, 128], [128, 128, 128], [64, 0, 0], [192, 0, 0],
                         [64, 128, 0], [192, 128, 0], [64, 0, 128], [192, 0, 128],
                         [64, 128, 128], [192, 128, 128], [0, 64, 0], [128, 64, 0],
                         [0, 192, 0], [128, 192, 0], [0, 64, 128]]
        # 辅助变量
        self.fnum = 0
        if transforms is None:
            normalize = T.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
            self.transforms = T.Compose([
                T.ToTensor(),
                normalize
            ])
        # 像素值(RGB)与类别label(0,1,3...)一一对应
        self.cm2lbl = np.zeros(256 ** 3)
        for i, cm in enumerate(self.colormap):
            self.cm2lbl[(cm[0] * 256 + cm[1]) * 256 + cm[2]] = i

        if train:
            txt_fname = root + "/ImageSets/Segmentation/train.txt"
        else:
            txt_fname = root + "/ImageSets/Segmentation/val.txt"
        with open(txt_fname, 'r') as f:
            images = f.read().split()
        imgs = [os.path.join(root, "JPEGImages", item + ".jpg") for item in images]
        labels = [os.path.join(root, "SegmentationClass", item + ".png") for item in images]
        self.imgs = self._filter(imgs)
        self.labels = self._filter(labels)
        if train:
            print("训练集：加载了 " + str(len(self.imgs)) + " 张图片和标签" + ",过滤了" + str(self.fnum) + "张图片")
        else:
            print("测试集：加载了 " + str(len(self.imgs)) + " 张图片和标签" + ",过滤了" + str(self.fnum) + "张图片")

    def _crop(self, data, label):
        """
        切割函数，默认都是从图片的左上角开始切割。切割后的图片宽是width,高是height
        data和label都是Image对象
        """
        box = (0, 0, self.width, self.height)
        data = data.crop(box)
        label = label.crop(box)
        return data, label

    def _image2label(self, im):
        data = np.array(im, dtype="int32")
        idx = (data[:, :, 0] * 256 + data[:, :, 1]) * 256 + data[:, :, 2]
        return np.array(self.cm2lbl[idx], dtype="int64")

    def _image_transforms(self, data, label):
        data, label = self._crop(data, label)
        data = self.transforms(data)
        label = self._image2label(label)
        label = torch.from_numpy(label)
        return data, label

    def _filter(self, imgs):
        img = []
        for im in imgs:
            if (Image.open(im).size[1] >= self.height and
                    Image.open(im).size[0] >= self.width):
                img.append(im)
            else:
                self.fnum = self.fnum + 1
        return img

    def __getitem__(self, index: int):
        img_path = self.imgs[index]
        label_path = self.labels[index]
        img = Image.open(img_path)
        label = Image.open(label_path).convert("RGB")
        img, label = self._image_transforms(img, label)
        return img, label

    def __len__(self):
        return len(self.imgs)


if __name__ == "__main__":
    root = "./VOCdevkit/VOC2007"
    height = 224
    width = 224
    voc_train = VOC_SEG(root, width, height, train=True)
    voc_test = VOC_SEG(root, width, height, train=False)

    # train_data = DataLoader(voc_train, batch_size=8, shuffle=True)
    # valid_data = DataLoader(voc_test, batch_size=8)
    for data, label in voc_train:
        print(data.shape)
        print(label.shape)
        break





'''enlarge_label_det1.0'''

import ctypes
import os
import sys

class enlarge_label_det:
    def __init__(self,dict_path,import_HALCONdll=True):
        # 如果电脑装了halcon且lisence在有效期，则import_HALCONdll=False,
        # 否则设置为True，导入dll，该操作开启默认从SegJJC的dll中获取
        self.dict_path=dict_path
        self.import_HALCONdll=import_HALCONdll

    def enlarge_label_area(self,labels, areamins, enlarged_dict_path):
        # 初始化 HALCON 环境
        if self.import_HALCONdll:
            self.setup_halcon_environment()
        import halcon as ha

        print("开始扩大...")

        # 读取字典
        dict_to_enlarge = ha.read_dict(self.dict_path, [], [])
        labellist_to_enlarge = ha.as_python_dict(dict_to_enlarge)

        # 获取 class_names 列表
        class_names = labellist_to_enlarge['class_names']

        # 对每个标签进行处理
        for label, areamin in zip(labels, areamins):
            if label not in class_names:
                print(f"Label '{label}' not found in class names.")
                continue

            # 获取标签对应的 labelid
            labelid_to_enlarge = class_names.index(label)

            # 获取 samples 列表
            samples = labellist_to_enlarge['samples']

            # 遍历 samples，查找匹配的 bbox_label_id 并获取对应的矩形坐标
            for sample in samples:
                bbox_label_ids = sample['bbox_label_id']

                # 遍历 bbox_label_ids 中所有等于 labelid_to_enlarge 的元素
                for i, bbox_label_id in enumerate(bbox_label_ids):
                    if bbox_label_id == labelid_to_enlarge:
                        # 获取矩形框的坐标
                        y1 = sample['bbox_row1'][i]
                        x1 = sample['bbox_col1'][i]
                        y2 = sample['bbox_row2'][i]
                        x2 = sample['bbox_col2'][i]

                        # 计算矩形面积
                        area = (y2 - y1) * (x2 - x1)

                        # 如果面积小于最小值，扩展矩形
                        if area < areamin:
                            # 计算扩展比例
                            scale_factor = (areamin / area) ** 0.5  # 保持矩形比例

                            # 计算新的矩形坐标，扩展至最小面积
                            center_y = (y1 + y2) / 2
                            center_x = (x1 + x2) / 2
                            height = y2 - y1
                            width = x2 - x1

                            new_height = height * scale_factor
                            new_width = width * scale_factor

                            # 计算扩展后的新的 y1, x1, y2, x2
                            y1 = int(center_y - new_height / 2)
                            x1 = int(center_x - new_width / 2)
                            y2 = int(center_y + new_height / 2)
                            x2 = int(center_x + new_width / 2)

                            # 更新矩形的值
                            sample['bbox_row1'][i] = y1
                            sample['bbox_col1'][i] = x1
                            sample['bbox_row2'][i] = y2
                            sample['bbox_col2'][i] = x2

        # 返回更新后的 labellist_enlarged
        labellist_enlarged = labellist_to_enlarge
        Dict_enlarged = ha.from_python_dict(labellist_enlarged)
        ha.write_dict(Dict_enlarged, enlarged_dict_path, [], [])
        print(f"已对{labels}等标签进行面积比例扩大!\n扩大后最小面积为{areamins}!\n并成功导出到:{enlarged_dict_path}")

    def setup_halcon_environment(self):
        """
        设置 HALCON 运行环境，包括动态加载 DLL 和许可证检查。
        """
        # 获取当前脚本路径
        current_dir = os.path.dirname(os.path.abspath(__file__))
        # DLL 文件夹路径（假设 DLL 文件夹在脚本的上一级目录）
        dll_path = os.path.join(os.path.dirname(current_dir), 'dll')

        # 动态添加 DLL 路径到 PATH 环境变量
        if dll_path not in os.environ['PATH']:
            os.environ['PATH'] = dll_path + os.pathsep + os.environ['PATH']

        # 手动加载 HALCON 主 DLL，确保动态库可用
        try:
            ctypes.windll.LoadLibrary(os.path.join(dll_path, "halcon.dll"))  # 替换为实际 DLL 文件名
            print("HALCON DLL 加载成功")
        except Exception as e:
            print(f"HALCON DLL 加载失败: {e}")
            sys.exit(1)  # 退出脚本，提示用户检查 DLL 配置


# ##使用##
# dict_path = r'E:\ALLvision\pycharmproject\pvk\infer\218\out\datajson\predictions.hdict'
# enlarged_dict_path = r'E:\ALLvision\pycharmproject\pvk\infer\218\out\datajson\predictions_enlarged.hdict'
# # labels = ['晕斑', '划痕']  # 输入多个标签
# # areamins = [100, 100]  # 对应每个标签的最小面积
# labels = ['晕斑']  # 输入多个标签
# areamins = [100]  # 对应每个标签的最小面积
# #实例化enlarge_label_det类
# enlargeit=enlarge_label_det(dict_path)
# # 调用 enlarge_label_area 函数
# enlargeit.enlarge_label_area(labels, areamins, enlarged_dict_path)




import os

import torch.utils.data as data
from PIL import Image,ImageDraw


import torch
# from torch.utils.data import DataLoader,Dataset,random_split
from torchvision import transforms
import cv2

class myDataset(data.Dataset):
    def __init__(self,dat_root,transforms=None):
        self.transforms = transforms
        self.imgsdir=dat_root+'/images'
        self.labelsdir = dat_root + '/labels'
        images_names=os.listdir(self.imgsdir)
        #到时可以实际情况修改是否加“_gt”
        self.masks = [os.path.join(self.labelsdir, x.split('.')[0] + '.png').replace('\\', '/') for x in images_names]
        self.images = [os.path.join(self.imgsdir, xi).replace('\\', '/') for xi in images_names]
    def __len__(self):
        return len(self.images)

    def __getitem__(self, index):
        img = Image.open(self.images[index]).convert('RGB')
        target = Image.open(self.masks[index])

        if self.transforms is not None:
            img, target = self.transforms(img, target)

        return img, target

    @staticmethod
    def collate_fn(batch):
        images, targets = list(zip(*batch))
        batched_imgs = cat_list(images, fill_value=0)
        batched_targets = cat_list(targets, fill_value=255)
        return batched_imgs, batched_targets

class myDataset_yaml(data.Dataset):
    def __init__(self,dat_root,transforms=None):
        self.transforms = transforms
        self.imgsdir=dat_root
        self.labelsdir = dat_root.split('/images')[0] + '/labels'+dat_root.split('/images')[1]
        images_names=os.listdir(self.imgsdir)
        #到时可以实际情况修改是否加“_gt”
        self.masks_txt = [os.path.join(self.labelsdir, x.split('.')[0] +'.txt').replace('\\', '/') for x in images_names]
        self.images = [os.path.join(self.imgsdir, xi).replace('\\', '/') for xi in images_names]
    def __len__(self):
        return len(self.images)

    def __getitem__(self, index):
        img = Image.open(self.images[index]).convert('RGB')
        target = txt2mask(img,self.masks_txt[index])

        if self.transforms is not None:
            img, target = self.transforms(img, target)

        return img, target

    @staticmethod
    def collate_fn(batch):
        images, targets = list(zip(*batch))
        batched_imgs = cat_list(images, fill_value=0)
        batched_targets = cat_list(targets, fill_value=255)
        return batched_imgs, batched_targets
def txt2mask(img,mask_txt):
    # 获取原始图像的尺寸
    width, height = img.size

    # 创建一个与图像相同尺寸的空白单通道（灰度）掩码图像
    mask = Image.new('L', (width, height), 0)
    draw = ImageDraw.Draw(mask)

    # 读取 .txt 文件
    with open(mask_txt, 'r') as file:
        lines = file.readlines()

    # 解析每一行
    for line in lines:
        parts = line.strip().split()
        label_index = int(parts[0])
        points = [float(p) for p in parts[1:]]

        # 将归一化的点坐标转换为图像坐标
        points = [(p * width, q * height) for p, q in zip(points[0::2], points[1::2])]

        # 绘制多边形并填充
        color = label_index + 1  # 使用标签索引+1作为灰度值
        draw.polygon(points, fill=color)
    # t=mask_txt.split("/")[-1].split(".txt")[0]
    # # 定义掩码图像的保存路径
    # mask_file_name = f"E:\\ALLvision\\pycharmproject\\yoloseg\\try\\826\\mask\\{t}.png"
    # # 保存掩码图像
    # mask.save(mask_file_name) # 保存掩码图像到文件
    return mask

class VOCSegmentation(data.Dataset):
    def __init__(self, voc_root, year="mydata", transforms=None, txt_name: str = "train.txt"):
        super(VOCSegmentation, self).__init__()
        assert year in ["2007", "2012","mydata"], "year must be in ['2007', '2012','mydata']"
        root = os.path.join(voc_root, "VOCdevkitMyself", f"VOC{year}")
        assert os.path.exists(root), "path '{}' does not exist.".format(root)
        image_dir = os.path.join(root, 'ImageDog')
        mask_dir = os.path.join(root, 'SegmentationClass')

        txt_path = os.path.join(root, "ImageSets", "Segmentation", txt_name)
        assert os.path.exists(txt_path), "file '{}' does not exist.".format(txt_path)
        with open(os.path.join(txt_path), "r") as f:
            file_names = [x.strip() for x in f.readlines() if len(x.strip()) > 0]

        self.images = [os.path.join(image_dir, x + ".jpg") for x in file_names]
        self.masks = [os.path.join(mask_dir, x + ".png") for x in file_names]
        assert (len(self.images) == len(self.masks))
        self.transforms = transforms

    def __getitem__(self, index):
        """
        Args:
            index (int): Index

        Returns:
            tuple: (image, target) where target is the image segmentation.
        """
        img = Image.open(self.images[index]).convert('RGB')
        target = Image.open(self.masks[index])

        if self.transforms is not None:
            img, target = self.transforms(img, target)

        return img, target

    def __len__(self):
        return len(self.images)

    @staticmethod
    def collate_fn(batch):
        images, targets = list(zip(*batch))
        batched_imgs = cat_list(images, fill_value=0)
        batched_targets = cat_list(targets, fill_value=255)
        return batched_imgs, batched_targets

def cat_list(images, fill_value=0):
    # 计算该batch数据中，channel, h, w的最大值
    max_size = tuple(max(s) for s in zip(*[img.shape for img in images]))
    batch_shape = (len(images),) + max_size
    batched_imgs = images[0].new(*batch_shape).fill_(fill_value)
    for img, pad_img in zip(images, batched_imgs):
        pad_img[..., :img.shape[-2], :img.shape[-1]].copy_(img)
    return batched_imgs

# dat_root="E:/ALLvision/pycharmproject/yoloseg/githubtry/diyFCN/try2_yaml/fcn/Data_seg/images/train"
# imgsdir = dat_root
# labelsdir = dat_root.split('/images')[0] + '/labels' + dat_root.split('/images')[1]
# images_names = os.listdir(imgsdir)
# masks_txt = [os.path.join(labelsdir, x.split('.')[0] + '.txt').replace('\\', '/') for x in images_names]
# images = [os.path.join(imgsdir, xi).replace('\\', '/') for xi in images_names]
# index=1
# img = Image.open(images[index]).convert('RGB')
# target = txt2mask(img, masks_txt[index])
# # 构造完整的文件路径
# target_path = "E:/ALLvision/pycharmproject/yoloseg/githubtry/diyFCN/try2_yaml/fcn/tryimgsave/1.png"
# # 保存 target 图像
# target.save(target_path)
# ss=Image.open(target_path)
# t=ss



import os
import torch

from src import fcn_resnet50
from train_utils import evaluate
from my_dataset import VOCSegmentation
import transforms as T


class SegmentationPresetEval:
    def __init__(self, base_size, mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)):
        self.transforms = T.Compose([
            T.RandomResize(base_size, base_size),
            T.ToTensor(),
            T.Normalize(mean=mean, std=std),
        ])

    def __call__(self, img, target):
        return self.transforms(img, target)


def main(args):
    device = torch.device(args.device if torch.cuda.is_available() else "cpu")
    assert os.path.exists(args.weights), f"weights {args.weights} not found."

    # segmentation nun_classes + background
    num_classes = args.num_classes + 1

    # VOCdevkit -> VOC2012 -> ImageSets -> Segmentation -> val.txt
    val_dataset = VOCSegmentation(args.data_path,
                                  year="2012",
                                  transforms=SegmentationPresetEval(520),
                                  txt_name="val.txt")

    num_workers = 8
    val_loader = torch.utils.data.DataLoader(val_dataset,
                                             batch_size=1,
                                             num_workers=num_workers,
                                             pin_memory=True,
                                             collate_fn=val_dataset.collate_fn)

    model = fcn_resnet50(aux=args.aux, num_classes=num_classes)
    model.load_state_dict(torch.load(args.weights, map_location=device)['model'])
    model.to(device)

    confmat = evaluate(model, val_loader, device=device, num_classes=num_classes)
    print(confmat)


def parse_args():
    import argparse
    parser = argparse.ArgumentParser(description="pytorch fcn training")

    parser.add_argument("--data-path", default="/data/", help="VOCdevkit root")
    parser.add_argument("--weights", default="./save_weights/model_29.pth")
    parser.add_argument("--num-classes", default=20, type=int)
    parser.add_argument("--aux", default=True, type=bool, help="auxilier loss")
    parser.add_argument("--device", default="cuda", help="training device")
    parser.add_argument('--print-freq', default=10, type=int, help='print frequency')

    args = parser.parse_args()

    return args


if __name__ == '__main__':
    args = parse_args()

    if not os.path.exists("./save_weights"):
        os.mkdir("./save_weights")

    main(args)



import json
import numpy as np
from PIL import Image

# 读取mask标签
target = Image.open("./2007_001288.png")
# 获取调色板
palette = target.getpalette()
palette = np.reshape(palette, (-1, 3)).tolist()
# 转换成字典子形式
pd = dict((i, color) for i, color in enumerate(palette))

json_str = json.dumps(pd)
with open("palette.json", "w") as f:
    f.write(json_str)

# target = np.array(target)
# print(target)



from collections import defaultdict, deque
import datetime
import time
import torch
import torch.distributed as dist

import errno
import os


class SmoothedValue(object):
    """Track a series of values and provide access to smoothed values over a
    window or the global series average.
    """

    def __init__(self, window_size=20, fmt=None):
        if fmt is None:
            fmt = "{value:.4f} ({global_avg:.4f})"
        self.deque = deque(maxlen=window_size)
        self.total = 0.0
        self.count = 0
        self.fmt = fmt

    def update(self, value, n=1):
        self.deque.append(value)
        self.count += n
        self.total += value * n

    def synchronize_between_processes(self):
        """
        Warning: does not synchronize the deque!
        """
        if not is_dist_avail_and_initialized():
            return
        t = torch.tensor([self.count, self.total], dtype=torch.float64, device='cuda')
        dist.barrier()
        dist.all_reduce(t)
        t = t.tolist()
        self.count = int(t[0])
        self.total = t[1]

    @property
    def median(self):
        d = torch.tensor(list(self.deque))
        return d.median().item()

    @property
    def avg(self):
        d = torch.tensor(list(self.deque), dtype=torch.float32)
        return d.mean().item()

    @property
    def global_avg(self):
        return self.total / self.count

    @property
    def max(self):
        return max(self.deque)

    @property
    def value(self):
        return self.deque[-1]

    def __str__(self):
        return self.fmt.format(
            median=self.median,
            avg=self.avg,
            global_avg=self.global_avg,
            max=self.max,
            value=self.value)


class ConfusionMatrix(object):
    def __init__(self, num_classes):
        self.num_classes = num_classes
        self.mat = None

    def update(self, a, b):
        n = self.num_classes
        if self.mat is None:
            # 创建混淆矩阵
            self.mat = torch.zeros((n, n), dtype=torch.int64, device=a.device)
        with torch.no_grad():
            # 寻找GT中为目标的像素索引
            k = (a >= 0) & (a < n)
            # 统计像素真实类别a[k]被预测成类别b[k]的个数(这里的做法很巧妙)
            inds = n * a[k].to(torch.int64) + b[k]
            self.mat += torch.bincount(inds, minlength=n**2).reshape(n, n)

    def reset(self):
        if self.mat is not None:
            self.mat.zero_()

    def compute(self):
        h = self.mat.float()
        # 计算全局预测准确率(混淆矩阵的对角线为预测正确的个数)
        acc_global = torch.diag(h).sum() / h.sum()
        # 计算每个类别的准确率
        acc = torch.diag(h) / h.sum(1)
        # 计算每个类别预测与真实目标的iou
        iu = torch.diag(h) / (h.sum(1) + h.sum(0) - torch.diag(h))
        return acc_global, acc, iu

    def reduce_from_all_processes(self):
        if not torch.distributed.is_available():
            return
        if not torch.distributed.is_initialized():
            return
        torch.distributed.barrier()
        torch.distributed.all_reduce(self.mat)

    def __str__(self):
        acc_global, acc, iu = self.compute()
        return (
            'global correct: {:.1f}\n'
            'average row correct: {}\n'
            'IoU: {}\n'
            'mean IoU: {:.1f}\n'
            'mean IoU(nobackground): {:.2f}').format(
                acc_global.item() * 100,
                ['{:.1f}'.format(i) for i in (acc * 100).tolist()],
                ['{:.1f}'.format(i) for i in (iu * 100).tolist()],
                iu.mean().item() * 100,
                iu[1:].mean().item() * 100)


class MetricLogger(object):
    def __init__(self, delimiter="\t"):
        self.meters = defaultdict(SmoothedValue)
        self.delimiter = delimiter

    def update(self, **kwargs):
        for k, v in kwargs.items():
            if isinstance(v, torch.Tensor):
                v = v.item()
            assert isinstance(v, (float, int))
            self.meters[k].update(v)

    def __getattr__(self, attr):
        if attr in self.meters:
            return self.meters[attr]
        if attr in self.__dict__:
            return self.__dict__[attr]
        raise AttributeError("'{}' object has no attribute '{}'".format(
            type(self).__name__, attr))

    def __str__(self):
        loss_str = []
        for name, meter in self.meters.items():
            loss_str.append(
                "{}: {}".format(name, str(meter))
            )
        return self.delimiter.join(loss_str)

    def synchronize_between_processes(self):
        for meter in self.meters.values():
            meter.synchronize_between_processes()

    def add_meter(self, name, meter):
        self.meters[name] = meter

    def log_every(self, iterable, print_freq, header=None):
        i = 0
        if not header:
            header = ''
        start_time = time.time()
        end = time.time()
        iter_time = SmoothedValue(fmt='{avg:.4f}')
        data_time = SmoothedValue(fmt='{avg:.4f}')
        space_fmt = ':' + str(len(str(len(iterable)))) + 'd'
        if torch.cuda.is_available():
            log_msg = self.delimiter.join([
                header,
                '[{0' + space_fmt + '}/{1}]',
                'eta: {eta}',
                '{meters}',
                'time: {time}',
                'data: {data}',
                'max mem: {memory:.0f}'
            ])
        else:
            log_msg = self.delimiter.join([
                header,
                '[{0' + space_fmt + '}/{1}]',
                'eta: {eta}',
                '{meters}',
                'time: {time}',
                'data: {data}'
            ])
        MB = 1024.0 * 1024.0
        for obj in iterable:
            data_time.update(time.time() - end)
            yield obj
            iter_time.update(time.time() - end)
            if i % print_freq == 0:
                eta_seconds = iter_time.global_avg * (len(iterable) - i)
                eta_string = str(datetime.timedelta(seconds=int(eta_seconds)))
                if torch.cuda.is_available():
                    print(log_msg.format(
                        i, len(iterable), eta=eta_string,
                        meters=str(self),
                        time=str(iter_time), data=str(data_time),
                        memory=torch.cuda.max_memory_allocated() / MB))
                else:
                    print(log_msg.format(
                        i, len(iterable), eta=eta_string,
                        meters=str(self),
                        time=str(iter_time), data=str(data_time)))
            i += 1
            end = time.time()
        total_time = time.time() - start_time
        total_time_str = str(datetime.timedelta(seconds=int(total_time)))
        print('{} Total time: {}'.format(header, total_time_str))


def mkdir(path):
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def setup_for_distributed(is_master):
    """
    This function disables printing when not in master process
    """
    import builtins as __builtin__
    builtin_print = __builtin__.print

    def print(*args, **kwargs):
        force = kwargs.pop('force', False)
        if is_master or force:
            builtin_print(*args, **kwargs)

    __builtin__.print = print


def is_dist_avail_and_initialized():
    if not dist.is_available():
        return False
    if not dist.is_initialized():
        return False
    return True


def get_world_size():
    if not is_dist_avail_and_initialized():
        return 1
    return dist.get_world_size()


def get_rank():
    if not is_dist_avail_and_initialized():
        return 0
    return dist.get_rank()


def is_main_process():
    return get_rank() == 0


def save_on_master(*args, **kwargs):
    if is_main_process():
        torch.save(*args, **kwargs)


def init_distributed_mode(args):
    if 'RANK' in os.environ and 'WORLD_SIZE' in os.environ:
        args.rank = int(os.environ["RANK"])
        args.world_size = int(os.environ['WORLD_SIZE'])
        args.gpu = int(os.environ['LOCAL_RANK'])
    elif 'SLURM_PROCID' in os.environ:
        args.rank = int(os.environ['SLURM_PROCID'])
        args.gpu = args.rank % torch.cuda.device_count()
    elif hasattr(args, "rank"):
        pass
    else:
        print('Not using distributed mode')
        args.distributed = False
        return

    args.distributed = True

    torch.cuda.set_device(args.gpu)
    args.dist_backend = 'nccl'
    print('| distributed init (rank {}): {}'.format(
        args.rank, args.dist_url), flush=True)
    torch.distributed.init_process_group(backend=args.dist_backend, init_method=args.dist_url,
                                         world_size=args.world_size, rank=args.rank)
    setup_for_distributed(args.rank == 0)



import time
import os
import datetime

import torch

from src import fcn_resnet50
from train_utils import train_one_epoch, evaluate, create_lr_scheduler, init_distributed_mode, save_on_master, mkdir
from my_dataset import VOCSegmentation
import transforms as T


class SegmentationPresetTrain:
    def __init__(self, base_size, crop_size, hflip_prob=0.5, mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)):
        min_size = int(0.5 * base_size)
        max_size = int(2.0 * base_size)

        trans = [T.RandomResize(min_size, max_size)]
        if hflip_prob > 0:
            trans.append(T.RandomHorizontalFlip(hflip_prob))
        trans.extend([
            T.RandomCrop(crop_size),
            T.ToTensor(),
            T.Normalize(mean=mean, std=std),
        ])
        self.transforms = T.Compose(trans)

    def __call__(self, img, target):
        return self.transforms(img, target)


class SegmentationPresetEval:
    def __init__(self, base_size, mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)):
        self.transforms = T.Compose([
            T.RandomResize(base_size, base_size),
            T.ToTensor(),
            T.Normalize(mean=mean, std=std),
        ])

    def __call__(self, img, target):
        return self.transforms(img, target)


def get_transform(train):
    base_size = 520
    crop_size = 480

    return SegmentationPresetTrain(base_size, crop_size) if train else SegmentationPresetEval(base_size)


def create_model(aux, num_classes):
    model = fcn_resnet50(aux=aux, num_classes=num_classes)
    weights_dict = torch.load("./fcn_resnet50_coco.pth", map_location='cpu')

    if num_classes != 21:
        # 官方提供的预训练权重是21类(包括背景)
        # 如果训练自己的数据集，将和类别相关的权重删除，防止权重shape不一致报错
        for k in list(weights_dict.keys()):
            if "classifier.4" in k:
                del weights_dict[k]

    missing_keys, unexpected_keys = model.load_state_dict(weights_dict, strict=False)
    if len(missing_keys) != 0 or len(unexpected_keys) != 0:
        print("missing_keys: ", missing_keys)
        print("unexpected_keys: ", unexpected_keys)

    return model


def main(args):
    init_distributed_mode(args)
    print(args)

    device = torch.device(args.device)
    # segmentation nun_classes + background
    num_classes = args.num_classes + 1

    # 用来保存coco_info的文件
    results_file = "results{}.txt".format(datetime.datetime.now().strftime("%Y%m%d-%H%M%S"))

    VOC_root = args.data_path
    # check voc root
    if os.path.exists(os.path.join(VOC_root, "VOCdevkit")) is False:
        raise FileNotFoundError("VOCdevkit dose not in path:'{}'.".format(VOC_root))

    # load train data set
    # VOCdevkit -> VOC2012 -> ImageSets -> Segmentation -> train.txt
    train_dataset = VOCSegmentation(args.data_path,
                                    year="2012",
                                    transforms=get_transform(train=True),
                                    txt_name="train.txt")
    # load validation data set
    # VOCdevkit -> VOC2012 -> ImageSets -> Segmentation -> val.txt
    val_dataset = VOCSegmentation(args.data_path,
                                  year="2012",
                                  transforms=get_transform(train=False),
                                  txt_name="val.txt")

    print("Creating data loaders")
    if args.distributed:
        train_sampler = torch.utils.data.distributed.DistributedSampler(train_dataset)
        test_sampler = torch.utils.data.distributed.DistributedSampler(val_dataset)
    else:
        train_sampler = torch.utils.data.RandomSampler(train_dataset)
        test_sampler = torch.utils.data.SequentialSampler(val_dataset)

    train_data_loader = torch.utils.data.DataLoader(
        train_dataset, batch_size=args.batch_size,
        sampler=train_sampler, num_workers=args.workers,
        collate_fn=train_dataset.collate_fn, drop_last=True)

    val_data_loader = torch.utils.data.DataLoader(
        val_dataset, batch_size=1,
        sampler=test_sampler, num_workers=args.workers,
        collate_fn=train_dataset.collate_fn)

    print("Creating model")
    # create model num_classes equal background + 20 classes
    model = create_model(aux=args.aux, num_classes=num_classes)
    model.to(device)

    if args.sync_bn:
        model = torch.nn.SyncBatchNorm.convert_sync_batchnorm(model)

    model_without_ddp = model
    if args.distributed:
        model = torch.nn.parallel.DistributedDataParallel(model, device_ids=[args.gpu])
        model_without_ddp = model.module

    params_to_optimize = [
        {"params": [p for p in model_without_ddp.backbone.parameters() if p.requires_grad]},
        {"params": [p for p in model_without_ddp.classifier.parameters() if p.requires_grad]},
    ]
    if args.aux:
        params = [p for p in model_without_ddp.aux_classifier.parameters() if p.requires_grad]
        params_to_optimize.append({"params": params, "lr": args.lr * 10})
    optimizer = torch.optim.SGD(
        params_to_optimize,
        lr=args.lr, momentum=args.momentum, weight_decay=args.weight_decay)

    scaler = torch.cuda.amp.GradScaler() if args.amp else None

    # 创建学习率更新策略，这里是每个step更新一次(不是每个epoch)
    lr_scheduler = create_lr_scheduler(optimizer, len(train_data_loader), args.epochs, warmup=True)

    # 如果传入resume参数，即上次训练的权重地址，则接着上次的参数训练
    if args.resume:
        # If map_location is missing, torch.load will first load the module to CPU
        # and then copy each parameter to where it was saved,
        # which would result in all processes on the same machine using the same set of devices.
        checkpoint = torch.load(args.resume, map_location='cpu')  # 读取之前保存的权重文件(包括优化器以及学习率策略)
        model_without_ddp.load_state_dict(checkpoint['model'])
        optimizer.load_state_dict(checkpoint['optimizer'])
        lr_scheduler.load_state_dict(checkpoint['lr_scheduler'])
        args.start_epoch = checkpoint['epoch'] + 1
        if args.amp:
            scaler.load_state_dict(checkpoint["scaler"])

    if args.test_only:
        confmat = evaluate(model, val_data_loader, device=device, num_classes=num_classes)
        val_info = str(confmat)
        print(val_info)
        return

    print("Start training")
    start_time = time.time()
    for epoch in range(args.start_epoch, args.epochs):
        if args.distributed:
            train_sampler.set_epoch(epoch)
        mean_loss, lr = train_one_epoch(model, optimizer, train_data_loader, device, epoch,
                                        lr_scheduler=lr_scheduler, print_freq=args.print_freq, scaler=scaler)

        confmat = evaluate(model, val_data_loader, device=device, num_classes=num_classes)
        val_info = str(confmat)
        print(val_info)

        # 只在主进程上进行写操作
        if args.rank in [-1, 0]:
            # write into txt
            with open(results_file, "a") as f:
                # 记录每个epoch对应的train_loss、lr以及验证集各指标
                train_info = f"[epoch: {epoch}]\n" \
                             f"train_loss: {mean_loss:.4f}\n" \
                             f"lr: {lr:.6f}\n"
                f.write(train_info + val_info + "\n\n")

        if args.output_dir:
            # 只在主节点上执行保存权重操作
            save_file = {'model': model_without_ddp.state_dict(),
                         'optimizer': optimizer.state_dict(),
                         'lr_scheduler': lr_scheduler.state_dict(),
                         'args': args,
                         'epoch': epoch}
            if args.amp:
                save_file["scaler"] = scaler.state_dict()
            save_on_master(save_file,
                           os.path.join(args.output_dir, 'model_{}.pth'.format(epoch)))

    total_time = time.time() - start_time
    total_time_str = str(datetime.timedelta(seconds=int(total_time)))
    print('Training time {}'.format(total_time_str))


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description=__doc__)

    # 训练文件的根目录(VOCdevkit)
    parser.add_argument('--data-path', default='/data/', help='dataset')
    # 训练设备类型
    parser.add_argument('--device', default='cuda', help='device')
    # 检测目标类别数(不包含背景)
    parser.add_argument('--num-classes', default=20, type=int, help='num_classes')
    # 每块GPU上的batch_size
    parser.add_argument('-b', '--batch-size', default=4, type=int,
                        help='images per gpu, the total batch size is $NGPU x batch_size')
    parser.add_argument("--aux", default=True, type=bool, help="auxilier loss")
    # 指定接着从哪个epoch数开始训练
    parser.add_argument('--start_epoch', default=0, type=int, help='start epoch')
    # 训练的总epoch数
    parser.add_argument('--epochs', default=20, type=int, metavar='N',
                        help='number of total epochs to run')
    # 是否使用同步BN(在多个GPU之间同步)，默认不开启，开启后训练速度会变慢
    parser.add_argument('--sync_bn', type=bool, default=False, help='whether using SyncBatchNorm')
    # 数据加载以及预处理的线程数
    parser.add_argument('-j', '--workers', default=4, type=int, metavar='N',
                        help='number of data loading workers (default: 4)')
    # 训练学习率，这里默认设置成0.0001，如果效果不好可以尝试加大学习率
    parser.add_argument('--lr', default=0.0001, type=float,
                        help='initial learning rate')
    # SGD的momentum参数
    parser.add_argument('--momentum', default=0.9, type=float, metavar='M',
                        help='momentum')
    # SGD的weight_decay参数
    parser.add_argument('--wd', '--weight-decay', default=1e-4, type=float,
                        metavar='W', help='weight decay (default: 1e-4)',
                        dest='weight_decay')
    # 训练过程打印信息的频率
    parser.add_argument('--print-freq', default=20, type=int, help='print frequency')
    # 文件保存地址
    parser.add_argument('--output-dir', default='./multi_train', help='path where to save')
    # 基于上次的训练结果接着训练
    parser.add_argument('--resume', default='', help='resume from checkpoint')
    # 不训练，仅测试
    parser.add_argument(
        "--test-only",
        dest="test_only",
        help="Only test the model",
        action="store_true",
    )

    # 分布式进程数
    parser.add_argument('--world-size', default=1, type=int,
                        help='number of distributed processes')
    parser.add_argument('--dist-url', default='env://', help='url used to set up distributed training')
    # Mixed precision training parameters
    parser.add_argument("--amp", default=False, type=bool,
                        help="Use torch.cuda.amp for mixed precision training")

    args = parser.parse_args()

    # 如果指定了保存文件地址，检查文件夹是否存在，若不存在，则创建
    if args.output_dir:
        mkdir(args.output_dir)

    main(args)



__version__ = "0.7.1"
from .model import EfficientNet, VALID_MODELS
from .utils import (
    GlobalParams,
    BlockArgs,
    BlockDecoder,
    efficientnet,
    get_model_params,
)
import SegJJC.EfficientNet.efficientnet_pytorch.transforms_cls as T_cls


import os
import json
import time
import torch
from torchvision import transforms
import numpy as np
from PIL import Image
import torchvision.models.segmentation as models
from pathlib import Path

from SegJJC.fcn.src import fcn_resnet50,fcn_resnet18,fcn_resnet34,deeplabv3_resnet18,deeplabv3_resnet34
try:
    from ultralytics.utils.plotting import colors
    ULTRALYTICS_COLORS_AVAILABLE = True
except ImportError:
    ULTRALYTICS_COLORS_AVAILABLE = False

class PredictSahi_fcn:
    def __init__(self, model_path,params):
        self.modelpath=model_path
        self.trueimgsize = params["inferimgsz"]
        self.testimg_dir = params['testimg']
        self.saveimg_dir = params['inferedimg']
        self.aux=False
        self.inferdevicehandle=params.get("inferdevicehandle",0)
        # if self.inferdevicehandle=='gpu':
        #     self.inferdevicehandle=0
        # 加载 palette
        self.palette = self.load_palette()
        self.yolocolors=True##是否使用yolo自带的掩膜color颜色盘，用于标记目标区域
        self.infer_format=params.get("infer_format","pt")
    def load_palette(self):
        """ 从 JSON 文件加载颜色映射 """
        json_path = os.path.join(os.path.dirname(__file__), "..", "fcn", "palette.json")
        if not os.path.exists(json_path):
            raise FileNotFoundError(f"颜色映射文件 {json_path} 不存在！")

        with open(json_path, "r") as f:
            return json.load(f)
    def time_synchronized(self):
        torch.cuda.synchronize() if torch.cuda.is_available() else None
        return time.time()

    def mask_to_color(self,mask,palette=None):
        """
        将语义分割得到的 mask（每个像素为类别ID）转换成彩色图
        """
        if not palette:
            palette=colors.palette
            h, w = mask.shape
            color_mask = np.zeros((h, w, 3), dtype=np.uint8)
            cls_id=0
            for color in palette:
                color_mask[mask == cls_id] = color
                cls_id+=1
            return color_mask
        else:
            """
            将语义分割得到的 mask（每个像素为类别ID）转换成彩色图
            """
            h, w = mask.shape
            color_mask = np.zeros((h, w, 3), dtype=np.uint8)
            for cls_id, color in palette.items():
                cls_id = int(cls_id)  # 将字符串类型的类别ID转换为整数
                color_mask[mask == cls_id] = color
            return color_mask

    def blend_mask_region(self,original_img, color_mask, mask, alpha=0.3):
        """
        仅对 mask 区域进行半透明叠加
          - original_img：原始RGB图像 (PIL Image)
          - color_mask：彩色mask (numpy数组，形状(H,W,3))
          - mask：语义分割结果 (numpy数组，像素值代表类别ID, 背景通常为0)
          - alpha：混合因子，在目标区域使用
        """
        # 转换为 numpy 数组（float32）
        original = np.array(original_img).astype(np.float32)
        overlay = color_mask.astype(np.float32)

        # 构造布尔 mask，将其沿通道重复3次，形状变为 (H, W, 3)
        mask_bool = np.repeat((mask > 0)[:, :, None], 3, axis=2)

        # 使用 np.where 只在 mask 区域混合，否则保持原图
        blended = np.where(mask_bool, original * (1 - alpha) + overlay * alpha, original)
        blended = np.clip(blended, 0, 255).astype(np.uint8)
        return Image.fromarray(blended)

    def process_image(self,model, device, img_path, output_path):
        # pt/pth模式推理
        if self.infer_format in ['pt', 'pth']:
            # 加载图像
            original_img = Image.open(img_path)
            if original_img.mode != 'RGB':
                original_img = original_img.convert('RGB')

            # 图像预处理
            data_transform = transforms.Compose([
                transforms.Resize(self.trueimgsize[0]),
                transforms.ToTensor(),
                transforms.Normalize(mean=(0.485, 0.456, 0.406),
                                     std=(0.229, 0.224, 0.225))
            ])
            img_tensor = data_transform(original_img)
            img_tensor = torch.unsqueeze(img_tensor, dim=0).to(device)

            model.eval()
            with torch.no_grad():
                t_start = self.time_synchronized()
                output = model(img_tensor)
                t_end = self.time_synchronized()
                print(f"Inference time for {os.path.basename(img_path)}: {t_end - t_start:.3f}s")

            # 假设模型输出 'out' 为分割结果
            prediction = output['out'].argmax(1).squeeze(0)
            mask = prediction.cpu().numpy().astype(np.uint8)

            # 将 mask 转为彩色图
            # 生成彩色 mask
            if ULTRALYTICS_COLORS_AVAILABLE and self.yolocolors:
                color_mask = self.mask_to_color(mask)
            else:
                color_mask = self.mask_to_color(mask, self.palette)
            # color_mask = self.mask_to_color(mask, self.palette)
            # 仅在 mask 区域进行半透明叠加
            blended_img = self.blend_mask_region(original_img, color_mask, mask, alpha=0.5)
            blended_img.save(output_path)
            print("Saved result to", output_path)
        # onnx模式推理
        if self.infer_format in ['onnx']:
            # 加载图像
            original_img = Image.open(img_path)
            if original_img.mode != 'RGB':
                original_img = original_img.convert('RGB')

            # 图像预处理
            data_transform = transforms.Compose([
                transforms.Resize(self.trueimgsize[0]),
                transforms.ToTensor(),
                transforms.Normalize(mean=(0.485, 0.456, 0.406),
                                     std=(0.229, 0.224, 0.225))
            ])
            img_tensor = data_transform(original_img).unsqueeze(0).numpy()  # 转换为 numpy

            # 获取 ONNX 输入名
            input_name = model.get_inputs()[0].name
            output_name = model.get_outputs()[0].name
            t_start = self.time_synchronized()
            output = model.run([output_name], {input_name: img_tensor})[0]
            t_end = self.time_synchronized()
            print(f"Inference time for {os.path.basename(img_path)}: {t_end - t_start:.3f}s")
            # 获取最大概率类别
            prediction = np.argmax(output, axis=1).squeeze(0).astype(np.uint8)

            # 将 mask 转为彩色图
            # 生成彩色 mask
            if ULTRALYTICS_COLORS_AVAILABLE and self.yolocolors:
                color_mask = self.mask_to_color(prediction)
            else:
                color_mask = self.mask_to_color(prediction, self.palette)
            # color_mask = self.mask_to_color(mask, self.palette)
            # 仅在 mask 区域进行半透明叠加
            blended_img = self.blend_mask_region(original_img, color_mask, prediction, alpha=0.5)
            blended_img.save(output_path)
            print("Saved result to", output_path)
        # openvino模式推理
        if self.infer_format in ['openvino']:
            # 加载图像
            original_img = Image.open(img_path)
            if original_img.mode != 'RGB':
                original_img = original_img.convert('RGB')

            # 图像预处理
            data_transform = transforms.Compose([
                transforms.Resize(self.trueimgsize[0]),
                transforms.ToTensor(),
                transforms.Normalize(mean=(0.485, 0.456, 0.406),
                                     std=(0.229, 0.224, 0.225))
            ])
            img_tensor = data_transform(original_img).unsqueeze(0).numpy()  # 转换为 numpy

            # 获取模型输入输出名称
            input_name = model.inputs[0].get_any_name()
            output_name = model.outputs[0].get_any_name()
            # 创建推理请求，并执行推理
            infer_request = model.create_infer_request()
            t_start = self.time_synchronized()
            results = infer_request.infer({input_name: img_tensor})
            t_end = self.time_synchronized()
            print(f"Inference time for {os.path.basename(img_path)}: {t_end - t_start:.3f}s")
            # 获取输出（假定输出形状为 (1, num_classes, 256, 256)）
            output = results[output_name]
            prediction = np.argmax(output, axis=1).squeeze(0).astype(np.uint8)

            # 将 mask 转为彩色图
            # 生成彩色 mask
            if ULTRALYTICS_COLORS_AVAILABLE and self.yolocolors:
                color_mask = self.mask_to_color(prediction)
            else:
                color_mask = self.mask_to_color(prediction, self.palette)
            # color_mask = self.mask_to_color(mask, self.palette)
            # 仅在 mask 区域进行半透明叠加
            blended_img = self.blend_mask_region(original_img, color_mask, prediction, alpha=0.5)
            blended_img.save(output_path)
            print("Saved result to", output_path)
    def predict_normal(self):
        #pt/pth模式推理
        if self.infer_format in ['pt','pth']:
            #获取所有设备
            # 获取所有 GPU 设备
            num_gpus = torch.cuda.device_count()
            gpu_list = [f"{i}: {torch.cuda.get_device_name(i)}" for i in range(num_gpus)]

            print(f"Available GPUs: {gpu_list}")
            # 配置设备
            if self.inferdevicehandle == 'cpu' :
                device = torch.device("cpu")
            elif self.inferdevicehandle == 'gpu':
                device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
            else:
                device = torch.device(f"cuda:{self.inferdevicehandle}" if torch.cuda.is_available() else "cpu")
            print("Using device:", device)
            # 创建模型（使用 FCN-ResNet50）
            model_dict = torch.load(self.modelpath, map_location='cpu')
            diymodel_dict = {
                "fcn_resnet50": fcn_resnet50,
                "fcn_resnet18": fcn_resnet18,
                "fcn_resnet34": fcn_resnet34,
                "deeplabv3_resnet18": deeplabv3_resnet18,
                "deeplabv3_resnet34": deeplabv3_resnet34,
            }

            model_arch = model_dict["model_type"]
            if model_arch in diymodel_dict:
                model = diymodel_dict[model_arch](aux=self.aux, num_classes=model_dict['num_classes'] + 1)
            else:
                model = models.__dict__[model_arch](pretrained=False, pretrained_backbone=False,
                                                          num_classes=model_dict['num_classes'] + 1,
                                                          aux_loss=self.aux)
            weights_dict = model_dict['model'].state_dict()
            for k in list(weights_dict.keys()):
                if "aux" in k:
                    del weights_dict[k]
            model.load_state_dict(weights_dict)
            model.to(device)
            # from pathlib import Path
            Path(self.saveimg_dir).mkdir(parents=True, exist_ok=True)
            # os.makedirs(self.saveimg_dir, exist_ok=True)#传统创建
            # # 遍历文件夹中的所有图像
            # for filename in os.listdir(self.testimg_dir):
            #     if filename.lower().endswith((".jpg", ".jpeg", ".png", ".bmp")):
            #         img_path = os.path.join(self.testimg_dir, filename)
            #         output_path = os.path.join(self.saveimg_dir, filename)
            #         print(f"Processing {img_path} ...")
            #         self.process_image(model, device, img_path, output_path)

            # 遍历 self.testimg_dir 下的所有子文件夹和图片
            for root, _, files in os.walk(self.testimg_dir):
                for filename in files:
                    if filename.lower().endswith((".jpg", ".jpeg", ".png", ".bmp")):
                        img_path = os.path.join(root, filename)  # 输入图像路径
                        # 计算相对路径
                        relative_path = os.path.relpath(root, self.testimg_dir)
                        # 生成输出目录，并保持相对路径结构
                        output_dir = os.path.join(self.saveimg_dir, relative_path)
                        os.makedirs(output_dir, exist_ok=True)
                        output_path = os.path.join(output_dir, filename)  # 输出路径

                        print(f"Processing {img_path} ...")
                        self.process_image(model, device, img_path, output_path)
        # onnx模式推理
        if self.infer_format in['onnx']:
            import onnxruntime as ort
            # 获取所有设备
            providers = ort.get_available_providers()
            print("Available ONNX Runtime providers:", providers)
            num_gpus = torch.cuda.device_count()
            gpu_list = [f"{i}: {torch.cuda.get_device_name(i)}" for i in range(num_gpus)]
            print(f"Available GPUs: {gpu_list}")
            #配置设备
            if self.inferdevicehandle == 'cpu':
                device = ['CPUExecutionProvider']
            elif self.inferdevicehandle == 'gpu':
                device = ['CUDAExecutionProvider' if torch.cuda.is_available() else 'CPUExecutionProvider']
            else:
                device = [('CUDAExecutionProvider',
                           {'device_id': int(self.inferdevicehandle)}) if torch.cuda.is_available() else 'CPUExecutionProvider']
                # device = torch.device(f"cuda:{self.inferdevicehandle}" if torch.cuda.is_available() else "cpu")
            print("Using device:", device)
            # 创建模型（使用 FCN-ResNet50）
            # 加载 ONNX 模型
            onnx_session = ort.InferenceSession(self.modelpath, providers=[
                'CUDAExecutionProvider' if torch.cuda.is_available() else 'CPUExecutionProvider'])

            # from pathlib import Path
            Path(self.saveimg_dir).mkdir(parents=True, exist_ok=True)
            # os.makedirs(self.saveimg_dir, exist_ok=True)#传统创建
            # # 遍历文件夹中的所有图像
            # for filename in os.listdir(self.testimg_dir):
            #     if filename.lower().endswith((".jpg", ".jpeg", ".png", ".bmp")):
            #         img_path = os.path.join(self.testimg_dir, filename)
            #         output_path = os.path.join(self.saveimg_dir, filename)
            #         print(f"Processing {img_path} ...")
            #         self.process_image(model, device, img_path, output_path)

            # 遍历 self.testimg_dir 下的所有子文件夹和图片
            for root, _, files in os.walk(self.testimg_dir):
                for filename in files:
                    if filename.lower().endswith((".jpg", ".jpeg", ".png", ".bmp")):
                        img_path = os.path.join(root, filename)  # 输入图像路径
                        # 计算相对路径
                        relative_path = os.path.relpath(root, self.testimg_dir)
                        # 生成输出目录，并保持相对路径结构
                        output_dir = os.path.join(self.saveimg_dir, relative_path)
                        os.makedirs(output_dir, exist_ok=True)
                        output_path = os.path.join(output_dir, filename)  # 输出路径

                        print(f"Processing {img_path} ...")
                        self.process_image(onnx_session, device, img_path, output_path)
        # openvino模式推理
        if self.infer_format in['openvino']:
            from openvino.runtime import Core
            #获取所有设备
            core = Core()
            devices = core.available_devices
            device_info = {device: core.get_property(device, "FULL_DEVICE_NAME") for device in devices}
            print("Available OpenVINO devices:", device_info)
            # 配置设备
            if self.inferdevicehandle == 'cpu':
                device = "CPU"
            elif self.inferdevicehandle == 'gpu':
                # device = "GPU"
                device = "GPU.0"
            else:
                device = f"GPU.{self.inferdevicehandle}" if torch.cuda.is_available() else "CPU"
                # device = torch.device(f"cuda:{self.inferdevicehandle}" if torch.cuda.is_available() else "cpu")
            print("Using device:", device)

            # 创建模型（使用 FCN-ResNet50）
            model = core.read_model(self.modelpath)
            compiled_model = core.compile_model(model, device)
            # from pathlib import Path
            Path(self.saveimg_dir).mkdir(parents=True, exist_ok=True)
            # os.makedirs(self.saveimg_dir, exist_ok=True)#传统创建
            # # 遍历文件夹中的所有图像
            # for filename in os.listdir(self.testimg_dir):
            #     if filename.lower().endswith((".jpg", ".jpeg", ".png", ".bmp")):
            #         img_path = os.path.join(self.testimg_dir, filename)
            #         output_path = os.path.join(self.saveimg_dir, filename)
            #         print(f"Processing {img_path} ...")
            #         self.process_image(model, device, img_path, output_path)

            # 遍历 self.testimg_dir 下的所有子文件夹和图片
            for root, _, files in os.walk(self.testimg_dir):
                for filename in files:
                    if filename.lower().endswith((".jpg", ".jpeg", ".png", ".bmp")):
                        img_path = os.path.join(root, filename)  # 输入图像路径
                        # 计算相对路径
                        relative_path = os.path.relpath(root, self.testimg_dir)
                        # 生成输出目录，并保持相对路径结构
                        output_dir = os.path.join(self.saveimg_dir, relative_path)
                        os.makedirs(output_dir, exist_ok=True)
                        output_path = os.path.join(output_dir, filename)  # 输出路径

                        print(f"Processing {img_path} ...")
                        self.process_image(compiled_model, device, img_path, output_path)
    # def predict_normal_onnx(self):
    #     # 配置设备
    #     if self.inferdevicehandle == 'cpu' :
    #         device = ['CPUExecutionProvider']
    #     elif self.inferdevicehandle == 'gpu':
    #         device = ['CUDAExecutionProvider' if torch.cuda.is_available() else 'CPUExecutionProvider']
    #     else:
    #         device = [('CUDAExecutionProvider', {'device_id': f"{self.inferdevicehandle}"}) if torch.cuda.is_available() else 'CPUExecutionProvider']
    #         # device = torch.device(f"cuda:{self.inferdevicehandle}" if torch.cuda.is_available() else "cpu")
    #     print("Using device:", device)
    #
    #     # 创建模型（使用 FCN-ResNet50）
    #     # 加载 ONNX 模型
    #     onnx_session = ort.InferenceSession(self.modelpath, providers=[
    #         'CUDAExecutionProvider' if torch.cuda.is_available() else 'CPUExecutionProvider'])
    #
    #     # from pathlib import Path
    #     Path(self.saveimg_dir).mkdir(parents=True, exist_ok=True)
    #     # os.makedirs(self.saveimg_dir, exist_ok=True)#传统创建
    #     # # 遍历文件夹中的所有图像
    #     # for filename in os.listdir(self.testimg_dir):
    #     #     if filename.lower().endswith((".jpg", ".jpeg", ".png", ".bmp")):
    #     #         img_path = os.path.join(self.testimg_dir, filename)
    #     #         output_path = os.path.join(self.saveimg_dir, filename)
    #     #         print(f"Processing {img_path} ...")
    #     #         self.process_image(model, device, img_path, output_path)
    #
    #     # 遍历 self.testimg_dir 下的所有子文件夹和图片
    #     for root, _, files in os.walk(self.testimg_dir):
    #         for filename in files:
    #             if filename.lower().endswith((".jpg", ".jpeg", ".png", ".bmp")):
    #                 img_path = os.path.join(root, filename)  # 输入图像路径
    #                 # 计算相对路径
    #                 relative_path = os.path.relpath(root, self.testimg_dir)
    #                 # 生成输出目录，并保持相对路径结构
    #                 output_dir = os.path.join(self.saveimg_dir, relative_path)
    #                 os.makedirs(output_dir, exist_ok=True)
    #                 output_path = os.path.join(output_dir, filename)  # 输出路径
    #
    #                 print(f"Processing {img_path} ...")
    #                 self.process_image(onnx_session, device, img_path, output_path)


import os
import yaml
from PIL import Image
import numpy as np

class ROIT_det:
    def __init__(self, yaml_path=None,roi_datadir=None,roi=None):
        self.yaml_path=yaml_path
        self.output_yaml_path=None
        self.Data_roi_dir=roi_datadir
        self.roi=roi

###获取切图区域内的标注框####
    def bbox_in_roi(self, bbox, roi_region):
        roi_x, roi_y, roi_w, roi_h = roi_region
        bbox_x1, bbox_y1, bbox_x2, bbox_y2 = bbox

        new_bbox_x1 = max(bbox_x1 - roi_x, 0)
        new_bbox_y1 = max(bbox_y1 - roi_y, 0)
        new_bbox_x2 = min(bbox_x2 - roi_x, roi_w)
        new_bbox_y2 = min(bbox_y2 - roi_y, roi_h)

        if new_bbox_x1 < new_bbox_x2 and new_bbox_y1 < new_bbox_y2:
            return new_bbox_x1, new_bbox_y1, new_bbox_x2, new_bbox_y2
        else:
            return None

    def convert_to_original(self, size, box):
        width, height = size
        x = box[0] * width
        y = box[1] * height
        w = box[2] * width
        h = box[3] * height

        x1 = x - w / 2
        y1 = y - h / 2
        x2 = x + w / 2
        y2 = y + h / 2

        return x1, y1, x2, y2
    def process_imgs(self, image_paths, output_dir):
        output_folder = os.path.join(output_dir, 'images')
        txt_outdir=os.path.join(output_dir,'labels')
        if not os.path.exists(output_folder):
            os.makedirs(output_folder)
        if not os.path.exists(txt_outdir):
            os.makedirs(txt_outdir)
        new_image_paths = []  # 用于存储新的图片路径

        for image_path in image_paths:
            ##获取图片格式
            imgtype = os.path.basename(image_path).split('.')[-1]
            txt_path = image_path.replace(f".{imgtype}", ".txt").replace("images", "labels")
            with open(txt_path, "r") as f:
                annotations = f.readlines()

            with Image.open(image_path) as img:
                img.load()
                width, height = img.size

                roi_x1, roi_y1, roi_x2, roi_y2 = self.roi
                roi_w=roi_x2-roi_x1
                roi_h=roi_y2 - roi_y1
                roi_region=[roi_x1, roi_y1, roi_w, roi_h]
                roi_img = img.crop((roi_x1, roi_y1, roi_x2 , roi_y2))

                roi_img_filename = os.path.join(output_folder, f"{os.path.basename(image_path).replace(f'.{imgtype}', '')}_roi.{imgtype}")
                roi_img.save(roi_img_filename)

                roi_txt_filename = os.path.join(txt_outdir, f"{os.path.basename(image_path).replace(f'.{imgtype}', '')}_roi.txt")
                with open(roi_txt_filename, "w") as roi_f:
                    has_bbox = False
                    for annotation in annotations:
                        elements = annotation.strip().split()
                        bbox_id = int(elements[0])
                        normalized_bbox = list(map(float, elements[1:]))

                        bbox = self.convert_to_original((width, height), normalized_bbox)

                        new_bbox = self.bbox_in_roi(bbox, roi_region)
                        ###归一化标注框
                        if new_bbox:
                            has_bbox = True
                            x_center = (new_bbox[0] + new_bbox[2]) / 2 / roi_w
                            y_center = (new_bbox[1] + new_bbox[3]) / 2 / roi_h
                            bbox_w = (new_bbox[2] - new_bbox[0]) / roi_w
                            bbox_h = (new_bbox[3] - new_bbox[1]) / roi_h

                            roi_f.write(f"{bbox_id} {x_center} {y_center} {bbox_w} {bbox_h}\n")

                    if not has_bbox:
                        roi_f.write("")

                new_image_paths.append(roi_img_filename)  # 添加新的图片路径

        return new_image_paths
    def update_yaml(self, train_paths, val_paths,test_paths):
        with open(self.yaml_path, 'r', encoding='utf-8') as file:
            data = yaml.safe_load(file)
        train_paths = train_paths.replace("\\", "/")
        val_paths = val_paths.replace("\\", "/")
        test_paths = test_paths.replace("\\", "/")
        data['train'] = train_paths
        data['val'] = val_paths
        data['test'] = test_paths
        with open(self.output_yaml_path, 'w', encoding='utf-8') as file:
            yaml.dump(data, file, allow_unicode=True)

    def save_image_paths(self, image_paths, output_txt_path):
        with open(output_txt_path, 'w') as f:
            for path in image_paths:
                f.write(f"{path}\n")
    def read_yaml(self):
        with open(self.yaml_path, 'r', encoding='utf-8') as file:
            config = yaml.safe_load(file)

        train_txt_path = config['train']
        val_txt_path = config['val']
        test_txt_path=train_txt_path.replace("train.txt",'test.txt')
        # 3. 读取 train 和 val 的文件路径
        train_image_paths = [line.strip() for line in open(train_txt_path, 'r')]
        val_image_paths = [line.strip() for line in open(val_txt_path, 'r')]
        # 检查 test.txt 文件是否存在
        if os.path.exists(test_txt_path):
            test_image_paths = [line.strip() for line in open(test_txt_path, 'r')]
        else:
            # 如果 test.txt 文件不存在，则返回空列表
            test_image_paths = []
            print(f"Warning: {test_txt_path} does not exist. Returning empty test image paths.")
        return train_image_paths, val_image_paths, test_image_paths


    def process_data(self):
        # 5. 处理 train 和 val 图片
        train_image_paths, val_image_paths, test_image_paths=self.read_yaml()
        if self.Data_roi_dir:
            Data_roi_dir=self.Data_roi_dir
        else:
        # 获取上层路径
            parent_dir = os.path.dirname(self.yaml_path)
            Data_roi_dir=os.path.join(parent_dir,'Data_det_roi')
        if not os.path.exists(Data_roi_dir):
            os.makedirs(Data_roi_dir)
        trainval_outdir=os.path.join(Data_roi_dir,'paper_data')
        if not os.path.exists(trainval_outdir):
            os.makedirs(trainval_outdir)
        new_train_paths = self.process_imgs(train_image_paths,Data_roi_dir)
        new_val_paths = self.process_imgs(val_image_paths, Data_roi_dir)
        #
        # ###和train、val一样使用切图切分，然后输出切图后的路径
        # new_test_paths = self.process_imgs(test_image_paths, Data_sahi_dir)
        # ###对test数据集不使用切图，直接给出原来一样的路径
        new_test_paths =test_image_paths
        #
        # 6. 保存新的 train.txt 和 val.txt 文件
        output_train_txt=os.path.join(trainval_outdir,'train.txt')
        output_val_txt=os.path.join(trainval_outdir,'val.txt')
        output_test_txt = os.path.join(trainval_outdir, 'test.txt')
        self.save_image_paths(new_train_paths, output_train_txt)
        self.save_image_paths(new_val_paths, output_val_txt)
        self.save_image_paths(new_test_paths, output_test_txt)
        # 7. 更新 YAML 文件
        self.output_yaml_path=os.path.join(Data_roi_dir,'mydata.yaml')
        self.update_yaml(output_train_txt, output_val_txt,output_test_txt)

    # ##predict_slice_imgs切分文件夹所有图片
    # def predict_slice_imgs(self, image_paths):
    #     sliced_images = []  # 存储切割后的图片数据
    #     for image_path in os.listdir(image_paths):
    #         # 获取图片格式
    #         sliced_image=[]
    #         image_path = os.path.join(image_paths,image_path)
    #         with Image.open(image_path) as img:
    #             width, height = img.size
    #             self.calculate_slice_regions(image_height=height, image_width=width,roi=self.roi)
    #
    #             for slice_region, index in self.slice_regions:
    #                 slice_x, slice_y, slice_w, slice_h = slice_region
    #                 slice_img = img.crop((slice_x, slice_y, slice_x + slice_w, slice_y + slice_h))
    #
    #                 # 将切割的图像保存为 numpy 数组（或直接传递到模型支持的格式）
    #                 slice_img_array = np.array(slice_img)
    #                 sliced_image.append(slice_img_array)
    #         sliced_images.append(sliced_image)
    #     # 返回所有切割后的图片列表
    #     return sliced_images
    #
    # ##predict_slice_oneimg切分单个文件图片
    # def predict_slice_oneimg(self, image_path_or_image):
    #     # 根据参数的类型来决定如何处理
    #     if isinstance(image_path_or_image, str):
    #         sliced_images = []  # 存储切割后的图片数据
    #         image_path=image_path_or_image
    #         with Image.open(image_path) as img:
    #             width, height = img.size
    #             self.calculate_slice_regions(image_height=height, image_width=width,roi=self.roi)
    #
    #             for slice_region, index in self.slice_regions:
    #                 slice_x, slice_y, slice_w, slice_h = slice_region
    #                 slice_img = img.crop((slice_x, slice_y, slice_x + slice_w, slice_y + slice_h))
    #
    #                 # 将切割的图像保存为 numpy 数组（或直接传递到模型支持的格式）
    #                 slice_img_array = np.array(slice_img)
    #                 sliced_images.append(slice_img_array)
    #         # 返回所有切割后的图片列表
    #         return sliced_images
    #     elif isinstance(image_path_or_image, Image.Image):
    #         sliced_images = []  # 存储切割后的图片数据
    #         img = image_path_or_image
    #         width, height = img.size
    #         self.calculate_slice_regions(image_height=height, image_width=width,roi=self.roi)
    #
    #         for slice_region, index in self.slice_regions:
    #             slice_x, slice_y, slice_w, slice_h = slice_region
    #             slice_img = img.crop((slice_x, slice_y, slice_x + slice_w, slice_y + slice_h))
    #
    #             # 将切割的图像保存为 numpy 数组（或直接传递到模型支持的格式）
    #             slice_img_array = np.array(slice_img)
    #             sliced_images.append(slice_img_array)
    #         # 返回所有切割后的图片列表
    #         return sliced_images
    #     else:
    #         raise TypeError("image_path_or_image 必须是字符串或 PIL.Image.Image")




import copy
import os
import torch
from torchvision.ops import nms
from PIL import Image
import numpy as np
from .Hsahi import HSAHI_det
from ultralytics import YOLO
import cv2
from ultralytics.engine.results import Results, Boxes
import json
import re
import ctypes
import sys

class PredictSahi_yolo:
    def __init__(self, model_path,params):
        # 初始化模型路径、文件路径、保存路径
        self.modelpath=model_path
        self.model = YOLO(model_path)
        self.trueimgsize=params["inferimgsz"]
        self.sahi = HSAHI_det(slice_width=params["inferimgsz"][0],
                              slice_height=params["inferimgsz"][1],
                              overlap_width_ratio=params.get("sahi_overlapratio",0.25),
                              overlap_height_ratio=params.get("sahi_overlapratio",0.25),
                              roi=params.get("sahi_roi",None))
        self.file_pathname = params['testimg']
        self.cur_path = params['inferedimg']
        self.inferedimg_onedir=params.get("inferedimg_onedir",True)
        self.predictions = []
        self.result_dict = []
        self.resultjson_filetype=params.get("resultjson_filetype","json")
        self.import_HALCONdll=params.get("import_HALCONdll",True)
        self.resultjson_path = params.get("resultjson_path", None)
        self.auto_label=params.get("auto_label",False)
        self.m_probability_threshold = params.get("prob_thresh",0.25)
        self.m_nms_threshold = params.get("nms_thresh",0.45)
        self.inferbatch=params.get("inferbatch",1)
        self.inferimgsz=params["inferimgsz"][0]
        self.inferdevicehandle=params.get("inferdevicehandle",0)
        if self.inferdevicehandle=='gpu':
            self.inferdevicehandle=0
        self.line_width=params.get("line_width",None)
        self.font_size=params.get("font_size",None)
        self.sahi_roi=params.get("sahi_roi",None)
        self.roi = params.get("roi", None)
        self.roi_padding=params.get("roi_padding", False)
        self.ifroiimg = params.get("ifroiimg", False)
        self.resize_before_roi=params.get("resize_before_roi", False)
        self.sahiinfer=params.get("sahiinfer",True)
        self.nmm_on_ori = params.get("nmm_on_ori", True)

    def process_detections_nms(self, it_sliceed, img):
        for q in range(len(it_sliceed)):
            # 获取当前切片的边界框、置信度和类别信息
            boxes = it_sliceed[q].boxes.xyxy  # 获取xyxy格式的边界框
            scores = it_sliceed[q].boxes.conf  # 置信度
            class_ids = it_sliceed[q].boxes.cls  # 类别ID

            # 过滤掉置信度低于阈值的框
            valid_indices = scores > self.m_probability_threshold
            boxes = boxes[valid_indices]
            scores = scores[valid_indices]
            class_ids = class_ids[valid_indices]

            # 进行 NMS
            if len(boxes) > 0:
                indices = nms(boxes, scores, self.m_nms_threshold)
                selected_boxes = boxes[indices]
                selected_scores = scores[indices]
                selected_class_ids = class_ids[indices]

                # 更新 it_sliceed[q].boxes
                new_data = torch.cat((selected_boxes, selected_scores.unsqueeze(1), selected_class_ids.unsqueeze(1)), dim=1)
                it_sliceed[q].boxes = Boxes(new_data, it_sliceed[q].boxes.orig_shape)

            imgori = np.array(img)
            it_sliceed[q].orig_img = imgori

        return it_sliceed

    # region greedy_nmm函数实现
    def nmm_calculate_box_union(self,bbox1, bbox2):
        """
        计算两个边界框的并集。
        :param bbox1: 第一个边界框 [x1, y1, x2, y2]
        :param bbox2: 第二个边界框 [x1, y1, x2, y2]
        :return: 并集边界框 [x1, y1, x2, y2]
        """
        union_bbox = np.zeros(4)
        union_bbox[0] = min(bbox1[0], bbox2[0])  # x1
        union_bbox[1] = min(bbox1[1], bbox2[1])  # y1
        union_bbox[2] = max(bbox1[2], bbox2[2])  # x2
        union_bbox[3] = max(bbox1[3], bbox2[3])  # y2
        return union_bbox

    def greedy_nmm(self, object_predictions, match_metric="IOU", match_threshold=0):
        """
        贪心NMS算法，用于合并重叠的边界框。
        :param object_predictions: 检测结果列表，每个元素包含 bbox, class_id, conf
        :param match_metric: 匹配标准（"IOU"）
        :param match_threshold: 匹配阈值
        :return: 合并后的检测结果列表
        """
        # 以置信度降序排序
        object_predictions = sorted(object_predictions, key=lambda x: x['conf'], reverse=True)
        selected = [False] * len(object_predictions)
        merged_predictions = []

        for i, pred in enumerate(object_predictions):
            if selected[i]:
                continue

            merged = pred.copy()  # 初始化合并的预测框
            selected[i] = True
            to_merge = [i]

            while to_merge:
                current_index = to_merge.pop(0)

                for j in range(len(object_predictions)):
                    if selected[j] or object_predictions[current_index]['class_id'] != object_predictions[j][
                        'class_id']:
                        continue

                    bbox1 = merged['bbox']
                    bbox2 = object_predictions[j]['bbox']

                    # 计算 IOU
                    xx1 = max(bbox1[0], bbox2[0])
                    yy1 = max(bbox1[1], bbox2[1])
                    xx2 = min(bbox1[2], bbox2[2])
                    yy2 = min(bbox1[3], bbox2[3])

                    w = max(0.0, xx2 - xx1)
                    h = max(0.0, yy2 - yy1)
                    inter = w * h

                    area1 = (bbox1[2] - bbox1[0]) * (bbox1[3] - bbox1[1])
                    area2 = (bbox2[2] - bbox2[0]) * (bbox2[3] - bbox2[1])
                    union_area = area1 + area2 - inter

                    iou = inter / union_area if union_area > 0 else 0.0

                    # 如果 IOU 大于阈值，则合并
                    if match_metric == "IOU" and iou > match_threshold:
                        selected[j] = True

                        # 合并框
                        union_bbox = self.nmm_calculate_box_union(bbox1, bbox2)
                        merged['bbox'] = union_bbox
                        merged['conf'] = max(merged['conf'], object_predictions[j]['conf'])
                        merged['class_id'] = merged['class_id'] if merged['conf'] > object_predictions[j]['conf'] else \
                        object_predictions[j]['class_id']

                        # 合并掩码（如果有）
                        if 'mask' in merged and 'mask' in object_predictions[j]:
                            # 如果掩码为None，初始化为空矩阵
                            if merged['mask'] is None:
                                merged['mask'] = np.zeros_like(object_predictions[j]['mask'])
                            if object_predictions[j]['mask'] is None:
                                object_predictions[j]['mask'] = np.zeros_like(merged['mask'])

                            # 合并掩码
                            merged['mask'] |= object_predictions[j]['mask']  # 假设掩码是二进制位图，使用位运算合并

                        to_merge.append(j)

            merged_predictions.append(merged)

        return merged_predictions
    def process_detections_with_nmm(self, it_sliceed,img):
        for q in range(len(it_sliceed)):
            boxes = it_sliceed[q].boxes.xyxy  # 获取经过 NMS 后的框
            scores = it_sliceed[q].boxes.conf  # 置信度
            class_ids = it_sliceed[q].boxes.cls  # 类别ID
            # 进行 NMM
            if len(boxes) > 0:
                # 准备 greedy_nmm 函数需要的输入格式
                object_predictions = []
                for i in range(len(boxes)):
                    object_predictions.append({
                        'bbox': boxes[i].cpu().numpy().tolist(),  # 转为list
                        'conf': scores[i].cpu().item(),  # 取值
                        'class_id': int(class_ids[i].cpu().item()),  # 转为整型
                        'mask': None  # 如果有掩码，这里可以加入掩码数据
                    })

                # 调用 greedy_nmm 函数进行合并
                merged_predictions = self.greedy_nmm(object_predictions, match_threshold=0)

                # 转回 YOLO 格式并更新结果
                merged_boxes = [pred['bbox'] for pred in merged_predictions]
                merged_scores = [pred['conf'] for pred in merged_predictions]
                merged_class_ids = [pred['class_id'] for pred in merged_predictions]

                # 更新 it_sliceed[q].boxes
                new_data = torch.cat((
                    torch.tensor(merged_boxes),
                    torch.tensor(merged_scores).unsqueeze(1),
                    torch.tensor(merged_class_ids).unsqueeze(1)
                ), dim=1)
                it_sliceed[q].boxes = Boxes(new_data, it_sliceed[q].boxes.orig_shape)

            imgori = np.array(img)
            it_sliceed[q].orig_img = imgori

        return it_sliceed

    def process_detections_with_nmsm(self, it_sliceed, img):
        for q in range(len(it_sliceed)):
            # 获取经过NMS的边界框、置信度和类别信息
            boxes = it_sliceed[q].boxes.xyxy  # 获取xyxy格式的边界框
            scores = it_sliceed[q].boxes.conf  # 置信度
            class_ids = it_sliceed[q].boxes.cls  # 类别ID

            # 过滤掉置信度低于阈值的框
            valid_indices = scores > self.m_probability_threshold
            boxes = boxes[valid_indices]
            scores = scores[valid_indices]
            class_ids = class_ids[valid_indices]

            # NMS 处理后的框
            if len(boxes) > 0:
                indices = nms(boxes, scores, self.m_nms_threshold)
                selected_boxes = boxes[indices]
                selected_scores = scores[indices]
                selected_class_ids = class_ids[indices]

                # 准备 greedy_nmm 函数需要的输入格式
                object_predictions = []
                for i in range(len(selected_boxes)):
                    object_predictions.append({
                        'bbox': selected_boxes[i].cpu().numpy().tolist(),  # 转为list
                        'conf': selected_scores[i].cpu().item(),  # 取值
                        'class_id': int(selected_class_ids[i].cpu().item()),  # 转为整型
                        'mask': None  # 如果有掩码，这里可以加入掩码数据
                    })

                # 调用 greedy_nmm 函数进行合并
                merged_predictions = self.greedy_nmm(object_predictions, match_threshold=0.5)

                # 转回 YOLO 格式并更新 it_sliceed[q].boxes
                merged_boxes = [pred['bbox'] for pred in merged_predictions]
                merged_scores = [pred['conf'] for pred in merged_predictions]
                merged_class_ids = [pred['class_id'] for pred in merged_predictions]

                new_data = torch.cat((
                    torch.tensor(merged_boxes),
                    torch.tensor(merged_scores).unsqueeze(1),
                    torch.tensor(merged_class_ids).unsqueeze(1)
                ), dim=1)
                it_sliceed[q].boxes = Boxes(new_data, it_sliceed[q].boxes.orig_shape)

            imgori = np.array(img)
            it_sliceed[q].orig_img = imgori

        return it_sliceed
    # endregion

    def merge_slices(self, it_sliced, img, shift_params,img_name):
        result_merged = []
        all_boxes_data = []

        for slice_idx, result in enumerate(it_sliced):
            boxes_data = result.boxes.data.clone()
            boxes_xyxy = boxes_data[:, :4].clone()

            # 对xyxy坐标进行偏移
            shift_x = shift_params[slice_idx][0][0]
            shift_y = shift_params[slice_idx][0][1]
            boxes_xyxy[:, 0] += shift_x  # x1
            boxes_xyxy[:, 1] += shift_y  # y1
            boxes_xyxy[:, 2] += shift_x  # x2
            boxes_xyxy[:, 3] += shift_y  # y2

            boxes_data[:, :4] = boxes_xyxy
            all_boxes_data.append(boxes_data)

        if all_boxes_data:
            merged_boxes_data = torch.cat(all_boxes_data, dim=0)
        else:
            merged_boxes_data = None

        oriimg = np.array(img)
        if merged_boxes_data is not None:
            result_merged.append(Results(oriimg, path=img_name, names=self.model.names, boxes=merged_boxes_data))

        return result_merged

    def recover_Roied(self, it_roied, img, shift_params,img_name):
        result_roied = []
        all_boxes_data = []

        for slice_idx, result in enumerate(it_roied):
            boxes_data = result.boxes.data.clone()
            boxes_xyxy = boxes_data[:, :4].clone()

            # 对xyxy坐标进行偏移
            shift_x = shift_params[0][0]
            shift_y = shift_params[0][1]
            shift_scale_x = shift_params[1][0]
            shift_scale_y = shift_params[1][1]
            boxes_xyxy[:, 0] += shift_x  # x1
            boxes_xyxy[:, 1] += shift_y  # y1
            boxes_xyxy[:, 2] += shift_x  # x2
            boxes_xyxy[:, 3] += shift_y  # y2
            ##scale如果有比例缩放，除以还原比例
            boxes_xyxy[:, 0] *= (1.0/shift_scale_x)  # x1
            boxes_xyxy[:, 1] *= (1.0/shift_scale_y)  # y1
            boxes_xyxy[:, 2] *= (1.0/shift_scale_x)  # x2
            boxes_xyxy[:, 3] *= (1.0/shift_scale_y)  # y2

            boxes_data[:, :4] = boxes_xyxy
            all_boxes_data.append(boxes_data)

        if all_boxes_data:
            roied_boxes_data = torch.cat(all_boxes_data, dim=0)
        else:
            roied_boxes_data = None

        oriimg = np.array(img)
        if roied_boxes_data is not None:
            result_roied.append(Results(oriimg, path=img_name, names=self.model.names, boxes=roied_boxes_data))

        return result_roied
    def recover_Roi_filled(self, it_roied, img, shift_params,img_name):
        result_roied = []
        all_boxes_data = []

        for slice_idx, result in enumerate(it_roied):
            boxes_data = result.boxes.data.clone()
            boxes_xyxy = boxes_data[:, :4].clone()

            # 对xyxy坐标进行偏移
            shift_scale_x = shift_params[1][0]
            shift_scale_y = shift_params[1][1]
            ##scale如果有比例缩放，除以还原比例
            boxes_xyxy[:, 0] *= (1.0/shift_scale_x)  # x1
            boxes_xyxy[:, 1] *= (1.0/shift_scale_y)  # y1
            boxes_xyxy[:, 2] *= (1.0/shift_scale_x)  # x2
            boxes_xyxy[:, 3] *= (1.0/shift_scale_y)  # y2

            boxes_data[:, :4] = boxes_xyxy
            all_boxes_data.append(boxes_data)

        if all_boxes_data:
            roied_boxes_data = torch.cat(all_boxes_data, dim=0)
        else:
            roied_boxes_data = None

        oriimg = np.array(img)
        if roied_boxes_data is not None:
            result_roied.append(Results(oriimg, path=img_name, names=self.model.names, boxes=roied_boxes_data))

        return result_roied
    def recover_Roi_nofilled(self, it_roied, img, shift_params,img_name):
        result_roied = []
        all_boxes_data = []

        for slice_idx, result in enumerate(it_roied):
            boxes_data = result.boxes.data.clone()
            boxes_xyxy = boxes_data[:, :4].clone()

            # 对xyxy坐标进行偏移
            shift_x = shift_params[0][0]
            shift_y = shift_params[0][1]
            shift_scale_x = shift_params[1][0]
            shift_scale_y = shift_params[1][1]
            ##scale如果有比例缩放，除以还原比例
            boxes_xyxy[:, 0] *= (1.0/shift_scale_x)  # x1
            boxes_xyxy[:, 1] *= (1.0/shift_scale_y)  # y1
            boxes_xyxy[:, 2] *= (1.0/shift_scale_x)  # x2
            boxes_xyxy[:, 3] *= (1.0/shift_scale_y)  # y2
            boxes_xyxy[:, 0] += shift_x  # x1
            boxes_xyxy[:, 1] += shift_y  # y1
            boxes_xyxy[:, 2] += shift_x  # x2
            boxes_xyxy[:, 3] += shift_y  # y2
            boxes_data[:, :4] = boxes_xyxy
            all_boxes_data.append(boxes_data)

        if all_boxes_data:
            roied_boxes_data = torch.cat(all_boxes_data, dim=0)
        else:
            roied_boxes_data = None

        oriimg = np.array(img)
        if roied_boxes_data is not None:
            result_roied.append(Results(oriimg, path=img_name, names=self.model.names, boxes=roied_boxes_data))

        return result_roied

    def roi_result_intersect(self, it_roied, img, roi_scaled, img_name):
        result_roied = []
        all_boxes_data = []
        # 提取ROI的坐标
        x1, y1, x2, y2 = roi_scaled
        for slice_idx, result in enumerate(it_roied):
            boxes_data = result.boxes.data.clone()
            boxes_xyxy = boxes_data[:, :4].clone()  # 提取前四列坐标 (x1, y1, x2, y2)
            scores_and_classes = boxes_data[:, 4:]  # 提取置信度和类别数据

            # 存储有效的检测框
            valid_boxes = []
            valid_scores_and_classes = []

            # 计算交集与ROI的交集部分
            for i in range(len(boxes_xyxy)):
                # 提取检测框的坐标
                det_x1, det_y1, det_x2, det_y2 = boxes_xyxy[i]

                # 计算交集区域
                intersect_x1 = max(det_x1, x1)
                intersect_y1 = max(det_y1, y1)
                intersect_x2 = min(det_x2, x2)
                intersect_y2 = min(det_y2, y2)

                # 如果交集区域有效，则保留这个检测框
                if intersect_x1 < intersect_x2 and intersect_y1 < intersect_y2:
                    # 更新检测框为交集部分
                    valid_boxes.append(torch.tensor([intersect_x1, intersect_y1, intersect_x2, intersect_y2]))
                    valid_scores_and_classes.append(scores_and_classes[i])  # 保留对应的置信度和类别

            # 如果有有效的检测框，更新 boxes_data
            if valid_boxes:
                valid_boxes = torch.stack(valid_boxes)
                valid_scores_and_classes = torch.stack(valid_scores_and_classes)
                # 确保 valid_boxes 和 valid_scores_and_classes 在同一设备上
                device = valid_scores_and_classes.device  # 获取 valid_boxes 所在设备
                valid_boxes= valid_boxes.to(device)  # 将 valid_scores_and_classes 转移到 valid_boxes 所在设备

                # 将有效的框坐标与对应的置信度、类别结合
                valid_boxes_data = torch.cat([valid_boxes, valid_scores_and_classes], dim=1)

                all_boxes_data.append(valid_boxes_data)
            else:
                # 创建空的张量 (0, 6) 和 (0, 4)
                empty_boxes = torch.empty(0, 6)  # 创建一个维度为 (0, 6) 的空张量
                device_empty=result.boxes.data.device
                empty_boxes=empty_boxes.to(device_empty)
                all_boxes_data.append(empty_boxes)
        # 合并所有有效的检测框
        if all_boxes_data:
            roied_boxes_data = torch.cat(all_boxes_data, dim=0)
        else:
            roied_boxes_data = None

        # 将图像转换为 NumPy 数组
        oriimg = np.array(img)

        # 如果有有效的检测框，保存到结果中
        if roied_boxes_data is not None:
            result_roied.append(Results(oriimg, path=img_name, names=self.model.names, boxes=roied_boxes_data))
        # 合并所有有效的检测框
        # roied_boxes_data = torch.cat(all_boxes_data, dim=0)
        # # 将图像转换为 NumPy 数组
        # oriimg = np.array(img)
        # # 如果有有效的检测框，保存到结果中
        # result_roied.append(Results(oriimg, path=img_name, names=self.model.names, boxes=roied_boxes_data))

        return result_roied
    def recover_Roi_intersect(self, it_roied, img, shift_params,img_name):
        result_roied = []
        all_boxes_data = []

        for slice_idx, result in enumerate(it_roied):
            boxes_data = result.boxes.data.clone()
            boxes_xyxy = boxes_data[:, :4].clone()

            # 对xyxy坐标进行偏移
            shift_scale_x = shift_params[0][0]
            shift_scale_y = shift_params[0][1]
            ##scale如果有比例缩放，除以还原比例
            boxes_xyxy[:, 0] *= (1.0/shift_scale_x)  # x1
            boxes_xyxy[:, 1] *= (1.0/shift_scale_y)  # y1
            boxes_xyxy[:, 2] *= (1.0/shift_scale_x)  # x2
            boxes_xyxy[:, 3] *= (1.0/shift_scale_y)  # y2
            boxes_data[:, :4] = boxes_xyxy
            all_boxes_data.append(boxes_data)

        if all_boxes_data:
            roied_boxes_data = torch.cat(all_boxes_data, dim=0)
        else:
            roied_boxes_data = None

        oriimg = np.array(img)
        if roied_boxes_data is not None:
            result_roied.append(Results(oriimg, path=img_name, names=self.model.names, boxes=roied_boxes_data))

        return result_roied

    # def predict_sahi(self):
    #     for filename in os.listdir(self.file_pathname):
    #         if filename.split('.')[-1] != 'txt':
    #             image_path = os.path.join(self.file_pathname, filename)
    #             with Image.open(image_path) as img:
    #                 sliced_images_one = self.sahi.predict_slice_oneimg(img)
    #                 result = self.model.predict(sliced_images_one, batch=self.inferbatch, imgsz=self.inferimgsz,
    #                                             device=self.inferdevicehandle)
    #                 merged_result = self.merge_slices(result, img, self.sahi.slice_regions, img_name=filename)
    #                 it_sliceed = self.process_detections_nms(merged_result, img)
    #                 self.predictions.append(it_sliceed)
    #
    #     self.save_results()
    #
    # def predict_oriimg(self):
    #     for filename in os.listdir(self.file_pathname):
    #         if filename.split('.')[-1] != 'txt':
    #             image_path = os.path.join(self.file_pathname, filename)
    #             with Image.open(image_path) as img:
    #                 result = self.model.predict(img, batch=self.inferbatch, imgsz=self.inferimgsz,
    #                                             device=self.inferdevicehandle)
    #                 self.predictions.append(result)
    #     self.save_results()

    # def predict_sahi(self,img,filename):
    #
    #     sliced_images_one = self.sahi.predict_slice_oneimg(img)
    #     result = self.model.predict(sliced_images_one, batch=self.inferbatch, imgsz=self.inferimgsz,
    #                                 device=self.inferdevicehandle)
    #     merged_result = self.merge_slices(result, img, self.sahi.slice_regions, img_name=filename)
    #     it_sliceed = self.process_detections_nms(merged_result, img)
    #     self.predictions.append(it_sliceed)
    #
    #     # self.save_results()
    def predict_sahi(self,img,filename):

        sliced_images_one = self.sahi.predict_slice_oneimg(img)
        dynamic_if=self.getmodel_inputbatch()
        if not dynamic_if and len(sliced_images_one)>1:
            result=[]
            for tt in range(0,len(sliced_images_one)):
                result_one = self.model.predict(sliced_images_one[tt], batch=self.inferbatch, imgsz=self.inferimgsz,
                                            device=self.inferdevicehandle)
                result.extend(result_one)
        else:
            result = self.model.predict(sliced_images_one, batch=self.inferbatch, imgsz=self.inferimgsz,
                                        device=self.inferdevicehandle)
        merged_result = self.merge_slices(result, img, self.sahi.slice_regions, img_name=filename)
        ###是否考虑将这nms、nmm只给切多图的，即len(sliced_images_one)>1，因为普通的可能无需后续操作###
        it_sliceed = self.process_detections_nms(merged_result, img)
        it_sliceed = self.process_detections_with_nmm(it_sliceed, img)
        ###是否考虑将这nms、nmm只给切多图的，即len(sliced_images_one)>1，因为普通的可能无需后续操作###
        self.predictions.append(it_sliceed)

        # self.save_results()

    def predict_oriimg(self,img):
        # 获取图像的宽度和高度
        img_width, img_height = img.size
        # 计算使用的图像尺寸
        effective_imgsz = self.inferimgsz
        if effective_imgsz >= img_width or effective_imgsz>= img_height:
            effective_imgsz = (img_width, img_height)
        else:
            effective_imgsz = (self.inferimgsz, self.inferimgsz)

        # 调用预测
        result = self.model.predict(img, batch=self.inferbatch, imgsz=effective_imgsz,
                                    device=self.inferdevicehandle)
        self.predictions.append(result)
        # result = self.model.predict(img, batch=self.inferbatch, imgsz=self.inferimgsz,
        #                             device=self.inferdevicehandle)
        # self.predictions.append(result)
        # # self.save_results()

    def predict_oriimg_noshai(self,img):
        # 调用预测
        result = self.model.predict(img, batch=self.inferbatch, imgsz=self.inferimgsz,
                                    device=self.inferdevicehandle)
        self.predictions.append(result)
        # result = self.model.predict(img, batch=self.inferbatch, imgsz=self.inferimgsz,
        #                             device=self.inferdevicehandle)
        # self.predictions.append(result)
        # # self.save_results()
    def predict_oriimg_noshai_nmm(self,img):
        # 调用预测
        result = self.model.predict(img, batch=self.inferbatch, imgsz=self.inferimgsz,
                                    device=self.inferdevicehandle)
        it_nmm = self.process_detections_with_nmm(result , img)
        self.predictions.append(it_nmm)
        # result = self.model.predict(img, batch=self.inferbatch, imgsz=self.inferimgsz,
        #                             device=self.inferdevicehandle)
        # self.predictions.append(result)
        # # self.save_results()

    def predict_sahi_roi(self,img,filename,roi):
        # self.sahiROI = HSAHI_det(slice_width=roi[2],
        #                       slice_height=roi[3],
        #                       overlap_width_ratio=0,
        #                       overlap_height_ratio=0,
        #                       roi=roi)
        self.sahiROI = HSAHI_det(slice_width=self.trueimgsize[0],
                              slice_height=self.trueimgsize[1],
                              overlap_width_ratio=0,
                              overlap_height_ratio=0,
                              roi=roi)
        #sahi_roi现在默认为先对图像使用roi然后再sahi
        sliced_images_one = self.sahiROI.predict_slice_oneimg(img)
        dynamic_if=self.getmodel_inputbatch()
        if not dynamic_if and len(sliced_images_one)>1:
            result=[]
            for tt in range(0,len(sliced_images_one)):
                result_one = self.model.predict(sliced_images_one[tt], batch=self.inferbatch, imgsz=self.inferimgsz,
                                            device=self.inferdevicehandle)
                result.extend(result_one)
        else:
            result = self.model.predict(sliced_images_one, batch=self.inferbatch, imgsz=self.inferimgsz,
                                        device=self.inferdevicehandle)
        merged_result = self.merge_slices(result, img, self.sahiROI.slice_regions, img_name=filename)
        ###是否考虑将这nms、nmm只给切多图的，即len(sliced_images_one)>1，因为普通的可能无需后续操作###
        it_sliceed = self.process_detections_nms(merged_result, img)
        it_sliceed = self.process_detections_with_nmm(it_sliceed, img)
        ###是否考虑将这nms、nmm只给切多图的，即len(sliced_images_one)>1，因为普通的可能无需后续操作###
        self.predictions.append(it_sliceed)

        # self.save_results()

    def predict_oriimg_roi_filled(self,img,filename,roi):
        # 确保图像是 RGB 模式
        if img.mode != "RGB":
            img = img.convert("RGB")
        # 转换 PIL 图像为 NumPy 数组
        img_np = np.array(img)
        # 提取 ROI 坐标
        x1, y1, x2, y2 = roi
        # 处理图像，将 ROI 外的部分填充为 0 值
        masked_img = np.zeros_like(img_np)  # 创建与 img 同大小的全 0 图像
        masked_img[y1:y2, x1:x2] = img_np[y1:y2, x1:x2]  # 仅保留 ROI 区域的像素值
        #shift-params计算
        shift_params=[]
        shift_params.append(roi)
        #只有roi，没有resize，比例则为1
        shift_params.append([1,1])
        # 调用预测
        result = self.model.predict(masked_img, batch=self.inferbatch, imgsz=self.inferimgsz,
                                    device=self.inferdevicehandle)
        #对结果做偏移(针对是边缘填充的)
        roied_result = self.recover_Roi_filled(result, img, shift_params, img_name=filename)
        ###是否考虑将这nms、nmm只给切多图的，即len(sliced_images_one)>1，因为普通的可能无需后续操作###
        it_roied = self.process_detections_nms(roied_result, img)
        it_roied = self.process_detections_with_nmm(it_roied, img)
        ###是否考虑将这nms、nmm只给切多图的，即len(sliced_images_one)>1，因为普通的可能无需后续操作###
        self.predictions.append(it_roied)

    # def predict_oriimg_resize_before_roi(self,img,filename,resize_imgsize,roi):
    #     # 确保图像是 RGB 模式
    #     if img.mode != "RGB":
    #         img = img.convert("RGB")
    #     # 调整图像大小为 1024x1024
    #     width,height=img.size
    #
    #     resized_img = img.resize((resize_imgsize[0], resize_imgsize[1]), Image.Resampling.LANCZOS)
    #     # 转换 PIL 图像为 NumPy 数组
    #     img_np = np.array(resized_img)
    #     # 提取 ROI 坐标
    #     x1, y1, x2, y2 = roi
    #     scale_roi_x=resize_imgsize[0]/width
    #     scale_roi_y=resize_imgsize[1]/height
    #     roi_x1=int(x1*scale_roi_x+0.5)
    #     roi_x2=int(x2*scale_roi_x+0.5)
    #     roi_y1=int(y1*scale_roi_y+0.5)
    #     roi_y2=int(y2*scale_roi_y+0.5)
    #     # 处理图像，将 ROI 外的部分填充为 0 值
    #     masked_img = np.zeros_like(img_np)  # 创建与 img 同大小的全 0 图像
    #     masked_img[roi_y1:roi_y2, roi_x1:roi_x2] = img_np[roi_y1:roi_y2, roi_x1:roi_x2]  # 仅保留 ROI 区域的像素值
    #     #shift-params计算
    #     scaled_roi=[roi_x1, roi_y1, roi_x2, roi_y2]
    #     shift_params=[]
    #     shift_params.append(scaled_roi)
    #     #只有roi，没有resize，比例则为1
    #     shift_params.append([scale_roi_x,scale_roi_y])
    #     # 调用预测
    #     result = self.model.predict(masked_img, batch=self.inferbatch, imgsz=self.inferimgsz,
    #                                 device=self.inferdevicehandle)
    #     #对结果做偏移(针对是边缘填充的)
    #     roied_result = self.recover_Roi_filled(result, img, shift_params, img_name=filename)
    #     ###是否考虑将这nms、nmm只给切多图的，即len(sliced_images_one)>1，因为普通的可能无需后续操作###
    #     it_roied = self.process_detections_nms(roied_result, img)
    #     it_roied = self.process_detections_with_nmm(it_roied, img)
    #     ###是否考虑将这nms、nmm只给切多图的，即len(sliced_images_one)>1，因为普通的可能无需后续操作###
    #     self.predictions.append(it_roied)
    def predict_oriimg_resize_before_roi(self,img,filename,resize_imgsize,roi,mode):
        # 确保图像是 RGB 模式
        if img.mode != "RGB":
            img = img.convert("RGB")
        # 调整图像大小为 1024x1024
        width,height=img.size
        if mode:
            resized_img = img.resize((resize_imgsize[0], resize_imgsize[1]), Image.Resampling.LANCZOS)
            # 转换 PIL 图像为 NumPy 数组
            img_np = np.array(resized_img)
            # 提取 ROI 坐标
            x1, y1, x2, y2 = roi
            scale_roi_x=resize_imgsize[0]/width
            scale_roi_y=resize_imgsize[1]/height
            roi_x1=int(x1*scale_roi_x+0.5)
            roi_x2=int(x2*scale_roi_x+0.5)
            roi_y1=int(y1*scale_roi_y+0.5)
            roi_y2=int(y2*scale_roi_y+0.5)
            # 处理图像，将 ROI 外的部分填充为 0 值
            masked_img = np.zeros_like(img_np)  # 创建与 img 同大小的全 0 图像
            masked_img[roi_y1:roi_y2, roi_x1:roi_x2] = img_np[roi_y1:roi_y2, roi_x1:roi_x2]  # 仅保留 ROI 区域的像素值
            #shift-params计算
            scaled_roi=[roi_x1, roi_y1, roi_x2, roi_y2]
            shift_params=[]
            shift_params.append(scaled_roi)
            #只有roi，没有resize，比例则为1
            shift_params.append([scale_roi_x,scale_roi_y])
            # 调用预测
            result = self.model.predict(masked_img, batch=self.inferbatch, imgsz=self.inferimgsz,
                                        device=self.inferdevicehandle)
            #对结果做偏移(针对是边缘填充的)
            roied_result = self.recover_Roi_filled(result, img, shift_params, img_name=filename)
            ###是否考虑将这nms、nmm只给切多图的，即len(sliced_images_one)>1，因为普通的可能无需后续操作###
            it_roied = self.process_detections_nms(roied_result, img)
            it_roied = self.process_detections_with_nmm(it_roied, img)
            ###是否考虑将这nms、nmm只给切多图的，即len(sliced_images_one)>1，因为普通的可能无需后续操作###
            self.predictions.append(it_roied)
        else:
            # 转换 PIL 图像为 NumPy 数组
            img_np = np.array(img)
            # 提取 ROI 坐标
            x1, y1, x2, y2 = roi
            x1 = max(0, min(x1, width - 1))
            x2 = max(0, min(x2, width - 1))
            y1 = max(0, min(y1, height - 1))
            y2 = max(0, min(y2, height - 1))
            # 裁剪图像到 ROI 区域
            roi_img = img.crop((x1, y1, x2, y2))  # 裁剪 ROI 区域
            longer_l=max((x2-x1),(y2-y1))
            scale_roi_x=resize_imgsize[0]/ longer_l
            scale_roi_y = resize_imgsize[1] / longer_l
            resized_roiw=int(((x2-x1)*scale_roi_x)+0.5)
            resized_roih=int(((y2-y1)*scale_roi_y)+0.5)
            resized_img=roi_img.resize((resized_roiw, resized_roih), Image.Resampling.LANCZOS)

            # shift-params计算
            scaled_roi = [x1, y1, x2, y2]
            shift_params = []
            shift_params.append(scaled_roi)
            # 比例
            shift_params.append([scale_roi_x, scale_roi_y])
            # 调用预测
            result = self.model.predict(resized_img, batch=self.inferbatch, imgsz=self.inferimgsz,
                                        device=self.inferdevicehandle)
            # 对结果做偏移(针对是边缘填充的)
            roied_result = self.recover_Roi_nofilled(result, img, shift_params, img_name=filename)
            ###是否考虑将这nms、nmm只给切多图的，即len(sliced_images_one)>1，因为普通的可能无需后续操作###
            it_roied = self.process_detections_nms(roied_result, img)
            it_roied = self.process_detections_with_nmm(it_roied, img)
            ###是否考虑将这nms、nmm只给切多图的，即len(sliced_images_one)>1，因为普通的可能无需后续操作###
            self.predictions.append(it_roied)
    # region PIL版predict_resizeimg_then_roi
    # def predict_resizeimg_then_roi(self,img,filename,resize_imgsize,roi):
    #     # 确保图像是 RGB 模式
    #     if img.mode != "RGB":
    #         img = img.convert("RGB")
    #     # 调整图像大小为 1024x1024
    #     width,height=img.size
    #
    #     resized_img = img.resize((resize_imgsize[0], resize_imgsize[1]), Image.Resampling.BILINEAR)
    #     # 转换 PIL 图像为 NumPy 数组
    #     img_np = np.array(resized_img)
    #     # 提取 ROI 坐标
    #     x1, y1, x2, y2 = roi
    #     scale_roi_x = resize_imgsize[0] / width
    #     scale_roi_y = resize_imgsize[1] / height
    #     roi_x1 = int(x1 * scale_roi_x + 0.5)
    #     roi_x2 = int(x2 * scale_roi_x + 0.5)
    #     roi_y1 = int(y1 * scale_roi_y + 0.5)
    #     roi_y2 = int(y2 * scale_roi_y + 0.5)
    #     scaled_roi = [roi_x1, roi_y1, roi_x2, roi_y2]
    #     # shift-params计算
    #     shift_params = []
    #     shift_params.append([scale_roi_x, scale_roi_y])
    #     # 调用预测
    #     result = self.model.predict(img_np, batch=self.inferbatch, imgsz=self.inferimgsz,
    #                                 device=self.inferdevicehandle)
    #     #对结果做偏移(针对是边缘填充的)
    #     roiinter_result = self.roi_result_intersect(result, img, scaled_roi, img_name=filename)
    #     roied_result = self.recover_Roi_intersect(roiinter_result, img, shift_params, img_name=filename)
    #     ###是否考虑将这nms、nmm只给切多图的，即len(sliced_images_one)>1，因为普通的可能无需后续操作###
    #     it_roied = self.process_detections_nms(roied_result, img)
    #     it_roied = self.process_detections_with_nmm(it_roied, img)
    #     ###是否考虑将这nms、nmm只给切多图的，即len(sliced_images_one)>1，因为普通的可能无需后续操作###
    #     self.predictions.append(it_roied)
    # endregion
    # region opencv版predict_resizeimg_then_roi
    def predict_resizeimg_then_roi(self, img, filename, resize_imgsize, roi):
        # 将 PIL 图像转换为 NumPy 数组
        img_np = np.array(img)

        # 确保图像是 RGB 模式，如果是其他模式先转换为 RGB
        if img.mode != "RGB":
            img = img.convert("RGB")
            img_np = np.array(img)

        # 获取原图的宽高
        height, width, _ = img_np.shape

        # 使用 OpenCV 的 resize 调整图像大小
        resized_img = cv2.resize(img_np, (resize_imgsize[0], resize_imgsize[1]), interpolation=cv2.INTER_LINEAR)

        # 提取 ROI 坐标
        x1, y1, x2, y2 = roi
        scale_roi_x = resize_imgsize[0] / width
        scale_roi_y = resize_imgsize[1] / height
        roi_x1 = int(x1 * scale_roi_x + 0.5)
        roi_x2 = int(x2 * scale_roi_x + 0.5)
        roi_y1 = int(y1 * scale_roi_y + 0.5)
        roi_y2 = int(y2 * scale_roi_y + 0.5)
        scaled_roi = [roi_x1, roi_y1, roi_x2, roi_y2]

        # shift-params计算
        shift_params = []
        shift_params.append([scale_roi_x, scale_roi_y])

        # 调用预测
        result = self.model.predict(resized_img, batch=self.inferbatch, imgsz=self.inferimgsz,
                                    device=self.inferdevicehandle)

        # 对结果做偏移（针对是边缘填充的）
        roiinter_result = self.roi_result_intersect(result, img, scaled_roi, img_name=filename)
        roied_result = self.recover_Roi_intersect(roiinter_result, img, shift_params, img_name=filename)

        # 是否考虑将这些 NMS 和 NMM 只给切割的多图操作
        it_roied = self.process_detections_nms(roied_result, img)
        it_roied = self.process_detections_with_nmm(it_roied, img)

        # 将结果保存到 predictions 中
        self.predictions.append(it_roied)
    # endregion
    def getmodel_inputbatch(self):
        ####判断输入batch是多少，然后确定每次训练个数###
        ###判断模型格式
        from ultralytics.nn.autobackend import AutoBackend
        ws = str(self.model.model[0] if isinstance(self.model.model, list) else self.model.model)
        nn_module = isinstance(ws, torch.nn.Module)
        (
            pt,
            jit,
            onnx,
            xml,
            engine,
            coreml,
            saved_model,
            pb,
            tflite,
            edgetpu,
            tfjs,
            paddle,
            ncnn,
            triton,
        ) = AutoBackend._model_type(ws)
        ###获取pt模型batch
        from ultralytics.nn.tasks import DetectionModel
        if  isinstance(self.model.model,DetectionModel):
            ifdynamic = True
        ###获取onnx模型batch
        if onnx:
            import onnxruntime
            providers = ["CUDAExecutionProvider", "CPUExecutionProvider"]
            session = onnxruntime.InferenceSession(ws, providers=providers)
            onnx_input_batch=session.get_inputs()[0].shape[0]##'batch'/1
            if onnx_input_batch=='batch':
                ifdynamic=True
            elif onnx_input_batch==1:
                ifdynamic=False
        ###获取openvino模型batch
        if xml:
            import openvino as ov
            from pathlib import Path
            w = Path(ws)
            if not w.is_file():  # if not *.xml
                w = next(w.glob("*.xml"))
            core = ov.Core()
            ov_model = core.read_model(model=str(w), weights=w.with_suffix(".bin"))
            ov_dynamic_batch = ov_model.dynamic##true/false
            ifdynamic=ov_dynamic_batch
        ###继续写，debug DoAi.py，定位到openvino的获取，json文件"infermodeldir"那个路径，train7为onnx dynamic=false，
        ####train8 为onnx dynamic=true，train9 为openvino dynamic=true ，train10为openvino dynamic=false
        ####判断输入batch是多少，然后确定每次训练个数###
        ###最后给出ifdynamic
        return  ifdynamic
    def save_results(self):
        original_filenames=os.listdir(self.file_pathname)
        for j, result in enumerate(self.predictions):
            # 获取原图的文件名
            original_filename = original_filenames[j] if j < len(original_filenames) else f"result_{j}"

            if isinstance(result, list):
                for res in result:
                    ann = res.plot(line_width=self.line_width,font_size=self.font_size)
                    if not os.path.exists(self.cur_path):
                        os.mkdir(self.cur_path)
                    # 根据原图文件名加后缀命名
                    output_filename = f"{os.path.splitext(original_filename)[0]}_result_{j}.jpg"
                    cv2.imwrite(os.path.join(self.cur_path, output_filename), ann)
            else:
                ann = result.plot(line_width=self.line_width,font_size=self.font_size)
                if not os.path.exists(self.cur_path):
                    os.mkdir(self.cur_path)
                # 根据原图文件名加后缀命名
                output_filename = f"{os.path.splitext(original_filename)[0]}_result_{j}.jpg"
                cv2.imwrite(os.path.join(self.cur_path, output_filename), ann)
    def save_results_oneout(self,test_image_paths):
        original_filenames=self.nooverlap_imgnames(test_image_paths)
        # original_filenames=os.listdir(self.file_pathname)
        for j, result in enumerate(self.predictions):
            # 获取原图的文件名
            original_filename = original_filenames[j] if j < len(original_filenames) else f"result_{j}"

            if isinstance(result, list):
                for res in result:
                    ann = res.plot(line_width=self.line_width,font_size=self.font_size)
                    if not os.path.exists(self.cur_path):
                        os.mkdir(self.cur_path)
                    # 根据原图文件名加后缀命名
                    output_filename = f"{os.path.splitext(original_filename)[0]}_result_{j}.jpg"
                    cv2.imwrite(os.path.join(self.cur_path, output_filename), ann)
            else:
                ann = result.plot(line_width=self.line_width,font_size=self.font_size)
                if not os.path.exists(self.cur_path):
                    os.mkdir(self.cur_path)
                # 根据原图文件名加后缀命名
                output_filename = f"{os.path.splitext(original_filename)[0]}_result_{j}.jpg"
                cv2.imwrite(os.path.join(self.cur_path, output_filename), ann)
        print(f"检测结果图已保存到：{self.cur_path}")
    def save_results_mulout(self,test_image_paths,non_common_files):
        for j, result in enumerate(self.predictions):
            # 获取原图的文件名
            original_filename = os.path.basename(test_image_paths[j]) if j < len(test_image_paths) else f"result_{j}"

            if isinstance(result, list):
                for res in result:
                    ann = res.plot(line_width=self.line_width,font_size=self.font_size)
                    if not os.path.exists(self.cur_path):
                        os.mkdir(self.cur_path)
                    # 根据原图文件名加后缀命名
                    output_filename = f"{os.path.splitext(original_filename)[0]}_result_{j}.jpg"
                    # 拼接保存路径
                    save_path = os.path.join(self.cur_path, os.path.dirname(non_common_files[j]))
                    # 确保保存路径存在
                    if not os.path.exists(save_path):
                        os.makedirs(save_path)
                    cv2.imwrite(os.path.join(save_path, output_filename), ann)
            else:
                ann = result.plot(line_width=self.line_width,font_size=self.font_size)
                if not os.path.exists(self.cur_path):
                    os.mkdir(self.cur_path)
                # 根据原图文件名加后缀命名
                output_filename = f"{os.path.splitext(original_filename)[0]}_result_{j}.jpg"
                # 拼接保存路径
                save_path = os.path.join(self.cur_path, os.path.dirname(non_common_files[j]))
                # 确保保存路径存在
                if not os.path.exists(save_path):
                    os.makedirs(save_path)
                cv2.imwrite(os.path.join(save_path, output_filename), ann)
        print(f"检测结果图已保存到：{self.cur_path}")
    # def predict_based_on_size(self):
    #     for filename in os.listdir(self.file_pathname):
    #         if filename.split('.')[-1] != 'txt':
    #             image_path = os.path.join(self.file_pathname, filename)
    #             with Image.open(image_path) as img:
    #                 width, height = img.size
    #                 # 判断图像大小
    #                 if width > self.trueimgsize[0] and height > self.trueimgsize[1]:
    #                     self.predict_sahi()
    #                 else:
    #                     self.predict_oriimg()
    def predict_sahiauto(self):
        for filename in os.listdir(self.file_pathname):
            if filename.split('.')[-1] != 'txt':
                image_path = os.path.join(self.file_pathname, filename)
                with Image.open(image_path) as img:
                    sliced_images_one = self.sahi.predict_slice_oneimg(img)
                    result = self.model.predict(sliced_images_one, batch=self.inferbatch, imgsz=self.inferimgsz,
                                                device=self.inferdevicehandle)
                    merged_result = self.merge_slices(result, img, self.sahi.slice_regions, img_name=filename)
                    it_sliceed = self.process_detections_nms(merged_result, img)
                    it_sliceed = self.process_detections_with_nmm(it_sliceed, img)
                    self.predictions.append(it_sliceed)

        self.save_results()
    # ###原版predict_based_on_size
    # def predict_based_on_size(self):
    #     for filename in os.listdir(self.file_pathname):
    #         if filename.split('.')[-1] != 'txt':
    #             image_path = os.path.join(self.file_pathname, filename)
    #             with Image.open(image_path) as img:
    #                 width, height = img.size
    #                 # 判断图像大小
    #                 if width > self.trueimgsize[0] or height > self.trueimgsize[1]:
    #                     self.predict_sahi(img,filename)
    #                 else:
    #                     self.predict_oriimg(img)
    #     self.save_results()
    ###重写predict_based_on_size，通过yaml里面的test路径读取图片
    def predict_based_on_size(self,yaml_datadir=None):
        if yaml_datadir:
            import  yaml
            with open(yaml_datadir, 'r', encoding='utf-8') as file:
                config = yaml.safe_load(file)
            if 'test' not in config or not config['test']:
                raise ValueError("'test' key is missing or its value is empty in the YAML file.Please remake mydata.yaml")
            test_txt_path = config['test']
            test_image_paths = [line.strip() for line in open(test_txt_path, 'r')]
        else:
            # test_image_paths=os.listdir(self.file_pathname)
            #对于一个文件夹里面好多分支，就不能简单使用os.listdir，而使用下面的
            test_image_paths=self.list_all_images_in_path(self.file_pathname)
        imgcommon_path=self.find_common_path(test_image_paths)
        image_id_dict=1
        non_common_files=[]
        for image_path in test_image_paths:
            # 标准化路径分隔符为 `/`
            standardized_path = "/".join(re.split(r"[\\/]", image_path))
            # 标准化公共路径分隔符为 `/`
            standardized_common_path = "/".join(re.split(r"[\\/]", imgcommon_path))

            # 获取非公共部分路径
            if standardized_path.startswith(standardized_common_path):
                non_common_part = standardized_path[len(standardized_common_path):].lstrip("/")
            else:
                non_common_part = standardized_path  # 如果路径不包含公共部分，直接使用原路径
            #读取尾缀为图片格式的图像
            filename=os.path.basename(image_path)
            imgfile_extension=(os.path.basename(filename).split('.')[-1]).lower()
            if imgfile_extension in {'jpg', 'jpeg', 'png', 'bmp', 'gif', 'tiff', 'webp'}:
                # image_path = os.path.join(self.file_pathname, filename)
                with Image.open(image_path) as img:
                    width, height = img.size
                    # 判断图像大小
                    if self.sahiinfer:
                        # if self.sahi_roi:
                        #     self.predict_sahi_roi(img, filename,self.sahi_roi)
                        if self.roi:
                            self.predict_sahi_roi(img, filename,self.roi)
                        else:
                            if width > self.trueimgsize[0] or height > self.trueimgsize[1]:
                                self.predict_sahi(img,filename)
                            else:
                                self.predict_oriimg(img)
                    else:
                        if self.roi:
                            # # if self.resize_before_roi:
                            # #     self.predict_oriimg_resize_before_roi(img,filename,self.trueimgsize,self.roi,self.roi_padding)
                            # # else:
                            # #     if self.roi_padding:
                            # #         self.predict_oriimg_roi_filled(img,filename, self.roi)
                            # #     else:
                            # #         self.predict_oriimg_resize_before_roi(img, filename, self.trueimgsize, self.roi,
                            # #                                               self.roi_padding)
                            if self.ifroiimg:
                                self.predict_oriimg_resize_before_roi(img, filename, self.trueimgsize, self.roi,
                                                                      self.roi_padding)
                            else:
                                self.predict_resizeimg_then_roi(img, filename, self.trueimgsize, self.roi)
                        else:
                            if self.nmm_on_ori:
                                self.predict_oriimg_noshai_nmm(img)
                            else:
                                self.predict_oriimg_noshai(img)
                    self.result_dict.append([self.predictions[image_id_dict-1],image_id_dict,non_common_part])
                    image_id_dict+=1
                    non_common_files.append(non_common_part)
        if self.auto_label:
            self.auto_labelit(imgcommon_path)
        ##考虑是否存在图片路径完全没有公共部分的！！！！
        if self.inferedimg_onedir:
            self.save_results_oneout(test_image_paths)
        else:
            self.save_results_mulout(test_image_paths,non_common_files)
    def auto_labelit(self,common_image_dir):
        # 存储JSON结果的列表
        file_pathnames=common_image_dir+"\\"
        json_results = {
            "class_ids": list(self.model.names.keys()),
            "class_names": list(self.model.names.values()),
            "image_dir": file_pathnames.replace("\\", "/"),
            "samples": []
        }
        for result,imageid_dict,filename_label in self.result_dict:
            # 提取并存储到JSON结果
            json_results["samples"].extend(self.extract_predictions(result, imageid_dict, filename_label))
        ##没有文件夹创建
        if not os.path.exists(self.resultjson_path):
            os.makedirs(self.resultjson_path)
        ##判断输出是哪种（hdict/json）
        if self.resultjson_filetype=='json':
            json_path = os.path.join(self.resultjson_path, 'predictions.json')
            with open(json_path, 'w', encoding='utf-8') as f:
                json.dump(json_results, f, ensure_ascii=False, indent=4)
        elif self.resultjson_filetype=='hdict':
            # 初始化 HALCON 环境
            if self.import_HALCONdll:
                self.setup_halcon_environment()
            import halcon as ha
            hdict_path = os.path.join(self.resultjson_path, 'predictions.hdict')
            Dict_ori = ha.from_python_dict(json_results)
            ha.write_dict(Dict_ori, hdict_path, [], [])
            print(f"检测标注hdict已成功导出到：{hdict_path}")

    def setup_halcon_environment(self):
        """
        设置 HALCON 运行环境，包括动态加载 DLL 和许可证检查。
        """
        # 获取当前脚本路径
        current_dir = os.path.dirname(os.path.abspath(__file__))
        # DLL 文件夹路径（假设 DLL 文件夹在脚本的上一级目录）
        dll_path = os.path.join(os.path.dirname(current_dir), 'dll')

        # 动态添加 DLL 路径到 PATH 环境变量
        if dll_path not in os.environ['PATH']:
            os.environ['PATH'] = dll_path + os.pathsep + os.environ['PATH']

        # 手动加载 HALCON 主 DLL，确保动态库可用
        try:
            ctypes.windll.LoadLibrary(os.path.join(dll_path, "halcon.dll"))  # 替换为实际 DLL 文件名
            print("HALCON DLL 加载成功")
        except Exception as e:
            print(f"HALCON DLL 加载失败: {e}")
            sys.exit(1)  # 退出脚本，提示用户检查 DLL 配置

    def extract_predictions(self, result, image_id, image_file_name):
        sample = []
        for det in result:
            # 将张量转换为列表
            xyxy_list = det.boxes.xyxy.tolist()
            # 初始化坐标列表
            col1_list = []
            row1_list = []
            col2_list = []
            row2_list = []
            # 遍历边界框坐标列表
            for box in xyxy_list:
                col1, row1, col2, row2 = box
                col1_list.append(col1)
                row1_list.append(row1)
                col2_list.append(col2)
                row2_list.append(row2)
            sample.append({
                "image_id": image_id,
                "image_file_name": image_file_name,
                "bbox_label_id": det.boxes.cls.to(torch.int).tolist(),
                "bbox_row1": row1_list,
                "bbox_col1": col1_list,
                "bbox_row2": row2_list,
                "bbox_col2": col2_list
            })
        return sample

    def list_all_images_in_path(self,base_path):
        # 定义支持的图片格式
        valid_extensions = ('.jpg', '.jpeg', '.png', '.bmp', '.tiff', '.gif')

        # 存储所有图片路径的列表
        all_image_paths = []

        # 遍历 base_path 目录及其子目录
        for root, dirs, files in os.walk(base_path):
            for file in files:
                # 检查文件扩展名是否为支持的图片格式
                if file.lower().endswith(valid_extensions):
                    # 构造完整文件路径
                    full_path = os.path.join(root, file)
                    all_image_paths.append(full_path)

        return all_image_paths
    def find_common_path(self,lines):
        # 分解每个路径为列表
        split_paths = [re.split(r"[\\/]", line) for line in lines]
        # 寻找公共部分
        common_parts = []
        for parts in zip(*split_paths):
            if all(part == parts[0] for part in parts):
                common_parts.append(parts[0])
            else:
                break
        # 拼接公共路径
        common_path = "\\".join(common_parts)
        return common_path

    def nooverlap_imgnames(self,test_image_paths):
        # 创建一个字典存储文件名和对应的路径信息
        name_to_path = {}
        final_names = []

        for image_path in test_image_paths:
            # 获取路径的各级部分
            path_parts = re.split(r"[\\/]", image_path.strip())
            # 最底层文件名
            current_name = path_parts[-1]
            # 检查文件名是否唯一
            temp_name = current_name
            index = -2  # 从倒数第二级目录开始
            while temp_name in name_to_path:
                # 如果重复，向前取上一级目录并加上前缀
                if abs(index) > len(path_parts):
                    raise ValueError(f"Cannot generate a unique name for: {image_path}")
                temp_name = f"{path_parts[index]}_{current_name}"
                index -= 1
            # 记录最终唯一名字
            name_to_path[temp_name] = image_path
            final_names.append(temp_name)

        # 返回图片名称列表，顺序与原列表一致
        return final_names
    # def save_results(self):
    #     for j, result in enumerate(self.predictions):
    #         if isinstance(result, list):
    #             for res in result:
    #                 ann = res.plot()
    #                 if not os.path.exists(self.cur_path):
    #                     os.mkdir(self.cur_path)
    #                 cv2.imwrite(os.path.join(self.cur_path, f"out{j}.jpg"), ann)
    #         else:
    #             ann = result.plot()
    #             if not os.path.exists(self.cur_path):
    #                 os.mkdir(self.cur_path)
    #             cv2.imwrite(os.path.join(self.cur_path, f"out{j}.jpg"), ann)

# 使用示例
# predictor = PredictSahi_yolo(
#     model_path=r"E:\ALLvision\pycharmproject\Efficientnet\segjjc_clsseg_try\sahi_preprocess\1014\try\traintry\runs\detect\train3\weights\best.pt",
#     yaml_path=r'E:\ALLvision\pycharmproject\Efficientnet\segjjc_clsseg_try\sahi_preprocess\1014\try\Data_pest\mydata2\mydata.yaml',
#     file_pathname=r"E:\ALLvision\pycharmproject\Efficientnet\segjjc_clsseg_try\sahi_preprocess\1014\predicttry\imgtest",
#     cur_path=r"E:\ALLvision\pycharmproject\Efficientnet\segjjc_clsseg_try\sahi_preprocess\1014\predicttry\imgsave"
# )
#
# predictor.predict_sahi()



from collections import OrderedDict

from typing import Dict

import torch
from torch import nn, Tensor
from torch.nn import functional as F
from .backbone import resnet50, resnet101


class IntermediateLayerGetter(nn.ModuleDict):
    """
    Module wrapper that returns intermediate layers from a model

    It has a strong assumption that the modules have been registered
    into the model in the same order as they are used.
    This means that one should **not** reuse the same nn.Module
    twice in the forward if you want this to work.

    Additionally, it is only able to query submodules that are directly
    assigned to the model. So if `model` is passed, `model.feature1` can
    be returned, but not `model.feature1.layer2`.

    Args:
        model (nn.Module): model on which we will extract the features
        return_layers (Dict[name, new_name]): a dict containing the names
            of the modules for which the activations will be returned as
            the key of the dict, and the value of the dict is the name
            of the returned activation (which the user can specify).
    """
    _version = 2
    __annotations__ = {
        "return_layers": Dict[str, str],
    }

    def __init__(self, model: nn.Module, return_layers: Dict[str, str]) -> None:
        if not set(return_layers).issubset([name for name, _ in model.named_children()]):
            raise ValueError("return_layers are not present in model")
        orig_return_layers = return_layers
        return_layers = {str(k): str(v) for k, v in return_layers.items()}

        # 重新构建backbone，将没有使用到的模块全部删掉
        layers = OrderedDict()
        for name, module in model.named_children():
            layers[name] = module
            if name in return_layers:
                del return_layers[name]
            if not return_layers:
                break

        super(IntermediateLayerGetter, self).__init__(layers)
        self.return_layers = orig_return_layers

    def forward(self, x: Tensor) -> Dict[str, Tensor]:
        out = OrderedDict()
        for name, module in self.items():
            x = module(x)
            if name in self.return_layers:
                out_name = self.return_layers[name]
                out[out_name] = x
        return out


class FCN(nn.Module):
    """
    Implements a Fully-Convolutional Network for semantic segmentation.

    Args:
        backbone (nn.Module): the network used to compute the features for the model.
            The backbone should return an OrderedDict[Tensor], with the key being
            "out" for the last feature map used, and "aux" if an auxiliary classifier
            is used.
        classifier (nn.Module): module that takes the "out" element returned from
            the backbone and returns a dense prediction.
        aux_classifier (nn.Module, optional): auxiliary classifier used during training
    """
    __constants__ = ['aux_classifier']

    def __init__(self, backbone, classifier, aux_classifier=None):
        super(FCN, self).__init__()
        self.backbone = backbone
        self.classifier = classifier
        self.aux_classifier = aux_classifier

    def forward(self, x: Tensor) -> Dict[str, Tensor]:
        input_shape = x.shape[-2:]
        # contract: features is a dict of tensors
        features = self.backbone(x)

        result = OrderedDict()
        x = features["out"]
        x = self.classifier(x)
        # 原论文中虽然使用的是ConvTranspose2d，但权重是冻结的，所以就是一个bilinear插值
        x = F.interpolate(x, size=input_shape, mode='bilinear', align_corners=False)
        result["out"] = x

        if self.aux_classifier is not None:
            x = features["aux"]
            x = self.aux_classifier(x)
            # 原论文中虽然使用的是ConvTranspose2d，但权重是冻结的，所以就是一个bilinear插值
            x = F.interpolate(x, size=input_shape, mode='bilinear', align_corners=False)
            result["aux"] = x

        return result


class FCNHead(nn.Sequential):
    def __init__(self, in_channels, channels):
        inter_channels = in_channels // 4
        layers = [
            nn.Conv2d(in_channels, inter_channels, 3, padding=1, bias=False),
            nn.BatchNorm2d(inter_channels),
            nn.ReLU(),
            nn.Dropout(0.1),
            nn.Conv2d(inter_channels, channels, 1)
        ]

        super(FCNHead, self).__init__(*layers)


def fcn_resnet50(aux, num_classes=21, pretrain_backbone=False):
    # 'resnet50_imagenet': 'https://download.pytorch.org/models/resnet50-0676ba61.pth'
    # 'fcn_resnet50_coco': 'https://download.pytorch.org/models/fcn_resnet50_coco-1167a1af.pth'
    backbone = resnet50(replace_stride_with_dilation=[False, True, True])

    if pretrain_backbone:
        # 载入resnet50 backbone预训练权重
        backbone.load_state_dict(torch.load("resnet50.pth", map_location='cpu'))

    out_inplanes = 2048
    aux_inplanes = 1024

    return_layers = {'layer4': 'out'}
    if aux:
        return_layers['layer3'] = 'aux'
    backbone = IntermediateLayerGetter(backbone, return_layers=return_layers)

    aux_classifier = None
    # why using aux: https://github.com/pytorch/vision/issues/4292
    if aux:
        aux_classifier = FCNHead(aux_inplanes, num_classes)

    classifier = FCNHead(out_inplanes, num_classes)

    model = FCN(backbone, classifier, aux_classifier)

    return model


def fcn_resnet101(aux, num_classes=21, pretrain_backbone=False):
    # 'resnet101_imagenet': 'https://download.pytorch.org/models/resnet101-63fe2227.pth'
    # 'fcn_resnet101_coco': 'https://download.pytorch.org/models/fcn_resnet101_coco-7ecb50ca.pth'
    backbone = resnet101(replace_stride_with_dilation=[False, True, True])

    if pretrain_backbone:
        # 载入resnet101 backbone预训练权重
        backbone.load_state_dict(torch.load("resnet101.pth", map_location='cpu'))

    out_inplanes = 2048
    aux_inplanes = 1024

    return_layers = {'layer4': 'out'}
    if aux:
        return_layers['layer3'] = 'aux'
    backbone = IntermediateLayerGetter(backbone, return_layers=return_layers)

    aux_classifier = None
    # why using aux: https://github.com/pytorch/vision/issues/4292
    if aux:
        aux_classifier = FCNHead(aux_inplanes, num_classes)

    classifier = FCNHead(out_inplanes, num_classes)

    model = FCN(backbone, classifier, aux_classifier)

    return model


def fcn_resnet18(aux, num_classes=21, pretrain_backbone=False):
    """
    FCN-ResNet18 模型
    Args:
        aux (bool): 是否使用辅助分类器
        num_classes (int): 类别数（包括背景）
        pretrain_backbone (bool): 是否加载 ImageNet 预训练权重
    Returns:
        FCN 模型
    """
    from torchvision.models import resnet18
    # 使用 ResNet18 作为 backbone，采用 replace_stride_with_dilation 调整特征分辨率
    backbone = resnet18(pretrained=pretrain_backbone, replace_stride_with_dilation=[False, False, False])

    # ResNet18 的最后输出通道数为 512；辅助分支（layer3）的输出通道数为 256
    out_inplanes = 512
    aux_inplanes = 256

    # 定义需要返回的层。默认返回 layer4 作为 "out"；如果使用辅助分类器，则同时返回 layer3 作为 "aux"
    return_layers = {'layer4': 'out'}
    if aux:
        return_layers['layer3'] = 'aux'

    backbone = IntermediateLayerGetter(backbone, return_layers=return_layers)

    aux_classifier = None
    if aux:
        aux_classifier = FCNHead(aux_inplanes, num_classes)
    classifier = FCNHead(out_inplanes, num_classes)

    model = FCN(backbone, classifier, aux_classifier)
    return model


def fcn_resnet34(aux, num_classes=21, pretrain_backbone=False):
    """
    FCN-ResNet34 模型
    Args:
        aux (bool): 是否使用辅助分类器
        num_classes (int): 类别数（包括背景）
        pretrain_backbone (bool): 是否加载 ImageNet 预训练权重
    Returns:
        FCN 模型
    """
    from torchvision.models import resnet34
    # 使用 ResNet34 作为 backbone，同样使用 replace_stride_with_dilation 调整特征分辨率
    backbone = resnet34(pretrained=pretrain_backbone, replace_stride_with_dilation=[False, False, False])

    # ResNet34 的输出通道与 ResNet18 相同：layer4 输出 512，layer3 输出 256
    out_inplanes = 512
    aux_inplanes = 256

    return_layers = {'layer4': 'out'}
    if aux:
        return_layers['layer3'] = 'aux'

    backbone = IntermediateLayerGetter(backbone, return_layers=return_layers)

    aux_classifier = None
    if aux:
        aux_classifier = FCNHead(aux_inplanes, num_classes)
    classifier = FCNHead(out_inplanes, num_classes)

    model = FCN(backbone, classifier, aux_classifier)
    return model
